#!/usr/bin/env python
from __future__ import absolute_import, division, print_function
import sys
import h5py
import numpy as np
from numpy import testing
import argparse
from NuRadioReco.utilities import units
import logging

file1 = sys.argv[1]
file2 = sys.argv[2]
print("Testing the files {} and {} for (almost) equality".format(file1, file2))

fin1 = h5py.File(file1, 'r')
fin2 = h5py.File(file2, 'r')

error = 0

accuracy = 0.0005


def test_equal_attributes(keys, fin1=fin1, fin2=fin2, error=error):
    for key in keys:
        try:
            testing.assert_equal(fin1.attrs[key], fin2.attrs[key])
        except AssertionError as e:
            print("\n attribute {} not almost equal".format(key))
            print(e)
            error = -1
    return error


def test_equal_station_keys(keys, fin1=fin1, fin2=fin2, error=error):
    for key in keys:
        try:
            testing.assert_equal(np.array(fin1['station_101'][key]), np.array(fin2['station_101'][key]))

        except AssertionError as e:
            print("\narray {} not almost equal".format(key))
            print("\Reference: {}, reconstruction: {}".format(fin2[key], fin1[key]))
            print(e)
            error = -1
    return error


def test_equal_keys(keys, fin1=fin1, fin2=fin2, error=error):
    for key in keys:
        try:
            testing.assert_equal(np.array(fin1[key]), np.array(fin2[key]))
        except AssertionError as e:
            print("\narray {} not almost equal".format(key))
            print("\Reference: {}, reconstruction: {}".format(fin2[key], fin1[key]))
            print(e)
            error = -1
    return error


def test_almost_equal_attributes(keys, fin1=fin1, fin2=fin2, error=error):
    for key in keys:
        arr1 = np.array(fin1.attrs[key])
        arr2 = np.array(fin2.attrs[key])
        max_diff = np.max(np.abs((arr1 - arr2) / arr2))
        if max_diff > accuracy:
            print('Reconstruction of {} does not agree with reference (error: {})'.format(key, max_diff))
            print("\n attribute {} not almost equal".format(key))
            error = -1
    return error


def test_almost_equal_station_keys(keys, fin1=fin1, fin2=fin2, error=error):
    for key in keys:
        arr1 = np.array(fin1['station_101'][key])
        arr2 = np.array(fin2['station_101'][key])
        for i in range(arr1.shape[0]):
            max_diff = np.max(np.abs((arr1[i] - arr2[i]) / arr2[i]))
#             print('Reconstruction of {} of event {} (relative error: {})'.format(key, i, np.abs((arr1[i] - arr2[i]) / arr2[i])))
            if max_diff > accuracy:
#                 print(arr1.shape)
                print('Reconstruction of {} of event {} does not agree with reference (relative error: {})'.format(key, i, max_diff))
                print("\n attribute {} not almost equal".format(key))
#                 print(np.abs((arr1[i] - arr2[i]) / arr2[i]))
#                 print(arr1[i])
                error = -1
    return error


def test_almost_equal_keys(keys, fin1=fin1, fin2=fin2, error=error):
    for key in keys:
        arr1 = np.array(fin1[key])
        arr2 = np.array(fin2[key])
        for i in range(arr1.shape[0]):
            max_diff = np.max(np.abs((arr1[i] - arr2[i]) / arr2[i]))
            if max_diff > accuracy:
                print('Reconstruction of {} of event {} does not agree with reference (error: {})'.format(key, i, max_diff))
                print("\n attribute {} not almost equal".format(key))
                error = -1
    return error

# Test those attributes that should be perfectly equal


attributes = [u'trigger_names',
 u'Tnoise',
 u'dt',
 u'bandwidth',
 u'n_samples',
 u'thetamin',
 u'zmax',
 u'zmin',
 u'thetamax',
 u'header',
 u'fiducial_zmax',
 u'fiducial_zmin',
 u'flavors',
 u'rmin',
 u'total_number_of_events',
 u'deposited',
 u'phimax',
 u'phimin',
 u'Emin',
 u'rmax',
 u'fiducial_rmax',
 u'Emax',
 u'fiducial_rmin',
 u'n_events']

error = test_equal_attributes(attributes, fin1=fin1, fin2=fin2, error=error)

# Test those attributes that should be numerically equal

attributes = [
 u'Vrms']

error = test_almost_equal_attributes(attributes, fin1=fin1, fin2=fin2, error=error)

# Test those station keys that should be perfectly equal

keys = [u'azimuths',
 u'energies',
 u'event_ids',
 u'flavors',
 u'inelasticity',
 u'interaction_type',
 u'multiple_triggers',
 u'n_interaction',
 u'triggered',
 u'xx',
 u'yy',
 u'zeniths',
 u'multiple_triggers',
 u'zz']
error = test_equal_keys(keys, fin1=fin1, fin2=fin2, error=error)

# Test those keys that should be perfectly equal

keys = [
u'ray_tracing_solution_type'
]
error = test_equal_station_keys(keys, fin1=fin1, fin2=fin2, error=error)

keys = [
 u'weights']

error = test_almost_equal_keys(keys, fin1=fin1, fin2=fin2, error=error)

keys = [
 u'SNRs',
 u'maximum_amplitudes',
 u'maximum_amplitudes_envelope',
u'polarization',
 u'ray_tracing_C0',
 u'launch_vectors',
 u'receive_vectors',
 u'travel_times',
 u'travel_distances',
 u'ray_tracing_C1',
 ]

error = test_almost_equal_station_keys(keys, fin1=fin1, fin2=fin2, error=error)

if error == -1:
    sys.exit(error)
else:
    print("The two files are (almost) identical.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import division, print_function
import numpy as np
from NuRadioReco.utilities import units, io_utilities
from scipy import interpolate as intp
from scipy import integrate as int
from scipy import constants
from matplotlib import pyplot as plt
from radiotools import coordinatesystems as cstrafo
from NuRadioReco.utilities.metaclasses import Singleton
import os
import copy
import logging
import six
logger = logging.getLogger("SignalGen.ARZ")
logging.basicConfig()

######################
######################
# This code is based on "J. Alvarez-Muniz, P. Hansen, A. Romero-Wolf, E. Zas in preparation" which is an extension of
# Alvarez-Muñiz, J., Romero-Wolf, A., & Zas, E. (2011). Practical and accurate calculations of Askaryan radiation. Physical Review D - Particles, Fields, Gravitation and Cosmology, 84(10). https://doi.org/10.1103/PhysRevD.84.103003
# to hadronic showers. Thanks a lot to Jaime Alvarez-Muñiz for providing us with this unpublished work!
#####################
#####################

# define constants
# x0 = 36.08 * units.g / units.cm**2  # radiation length g cm^-2
rho = 0.924 * units.g / units.cm ** 3  # density g cm^-3
xmu = 12.566370e-7 * units.newton / units.ampere ** 2
c = 2.99792458e8 * units.m / units.s
# e = 1.602177e-19 * units.coulomb


def thetaprime_to_theta(thetaprime, xmax, R):
    """
    convertes a viewing angle relative to the shower maximum to a viewing angle relative to the start of the shower.
    """
    L = xmax / rho
    return thetaprime - np.arcsin((L * np.sin(np.pi - thetaprime)) / R)


def theta_to_thetaprime(theta, xmax, R):
    """
    converts a viewing angle relative to the start of the shower to a viewing angle relative to the shower maximum
    """
    L = xmax / rho
    b = R * np.sin(theta)
    a = R * np.cos(theta) - L
    return np.arctan2(b, a)


@six.add_metaclass(Singleton)
class ARZ(object):

    def __init__(self, seed=1234, interp_factor=1, interp_factor2=100, library=None,
                 arz_version='ARZ2020'):
        logger.warning("setting seed to {}".format(seed, interp_factor))
        self._random_generator = np.random.RandomState(seed)
        self._interp_factor = interp_factor
        self._interp_factor2 = interp_factor2
        self._random_numbers = {}
        self._version = (1, 2)
        # # load shower library into memory
        if(library is None):
            library = os.path.join(os.path.dirname(__file__), "shower_library/library_v{:d}.{:d}.pkl".format(*self._version))
        else:
            if(not os.path.exists(library)):
                logger.error("user specified shower library {} not found.".format(library))
                raise FileNotFoundError("user specified shower library {} not found.".format(library))
        self.__check_and_get_library()
        self.__set_model_parameters(arz_version)

        logger.warning("loading shower library ({}) into memory".format(library))
        self._library = io_utilities.read_pickle(library)

    def __check_and_get_library(self):
        """
        checks if shower library exists and is up to date by comparing the sha1sum. If the library does not exist
        or changes on the server, a new library will be downloaded.
        """
        path = os.path.join(os.path.dirname(__file__), "shower_library/library_v{:d}.{:d}.pkl".format(*self._version))

        download_file = False
        if(not os.path.exists(path)):
            logger.warning("shower library version {} does not exist on the local file system yet. It will be downloaded to {}".format(self._version, path))
            download_file = True

        if(os.path.exists(path)):
            BUF_SIZE = 65536 * 2 ** 4  # lets read stuff in 64kb chunks!
            import hashlib
            import json
            sha1 = hashlib.sha1()
            with open(path, 'rb') as f:
                while True:
                    data = f.read(BUF_SIZE)
                    if not data:
                        break
                    sha1.update(data)

            shower_directory = os.path.join(os.path.dirname(__file__), "shower_library/")
            with open(os.path.join(shower_directory, 'shower_lib_hash.json'), 'r') as fin:
                lib_hashs = json.load(fin)
                if("{:d}.{:d}".format(*self._version) in lib_hashs.keys()):
                    if(sha1.hexdigest() != lib_hashs["{:d}.{:d}".format(*self._version)]):
                        logger.warning("shower library {} has changed on the server. downloading newest version...".format(self._version))
                        download_file = True
                else:
                    logger.warning("no hash sum of {} available, skipping up-to-date check".format(os.path.basename(path)))
        if not download_file:
            return True
        else:
            import requests
            URL = 'http://arianna.ps.uci.edu/~arianna/data/ce_shower_library/library_v{:d}.{:d}.pkl'.format(*self._version)

            logger.info("downloading shower library {} from {}. This can take a while...".format(self._version, URL))
            r = requests.get(URL)
            if (r.status_code != requests.codes.ok):
                logger.error("error in download of antenna model")
                raise IOError("error in download of antenna model")
            with open(path, "wb") as code:
                code.write(r.content)
            logger.info("...download finished.")

    def __set_model_parameters(self, arz_version='ARZ2020'):
        """
        Sets the parameters for the form factor
        """
        if (arz_version == 'ARZ2019'):
            # Refit of ZHAireS results => factor 0.88 in Af_e
            self._Af_e = -4.5e-14 * 0.88 * units.V * units.s
            self._t0_e_pos = 0.057 * units.ns
            self._freq_e_pos = 2.87 / units.ns
            self._exp_e_pos = -3.00
            self._t0_e_neg = 0.030 * units.ns
            self._freq_e_neg = 3.05 / units.ns
            self._exp_e_neg = -3.50

            self._Af_p = -3.2e-14 * units.V * units.s
            self._t0_p_pos = 0.065 * units.ns
            self._freq_p_pos = 3.00 / units.ns
            self._exp_p_pos = -2.65
            self._t0_p_neg = 0.043 * units.ns
            self._freq_p_neg = 2.92 / units.ns
            self._exp_p_neg = -3.21
            self._include_em_factor = False
        elif (arz_version == 'ARZ2020'):
            self._Af_e = -4.445e-14 * units.V * units.s
            self._t0_e_pos = 0.0348 * units.ns
            self._freq_e_pos = 2.298 / units.ns
            self._exp_e_pos = -3.588
            self._t0_e_neg = 0.0203 * units.ns
            self._freq_e_neg = 2.616 / units.ns
            self._exp_e_neg = -4.043

            self._Af_p = -4.071e-14 * units.V * units.s
            self._t0_p_pos = 0.0391 * units.ns
            self._freq_p_pos = 2.338 / units.ns
            self._exp_p_pos = -3.320
            self._t0_p_neg = 0.0234 * units.ns
            self._freq_p_neg = 2.686 / units.ns
            self._exp_p_neg = -3.687
            self._include_em_factor = True
        else:
            raise ValueError('ARZ version does not exist. Please choose ARZ2019 or ARZ2020.')

    def em_fraction(self, energy):
        """
        Returns the energy fraction carried by the electromagnetic component of
        a hadronic shower
        """

        if not self._include_em_factor:
            return 1

        epsilon = np.log10(energy / units.eV)
        f_epsilon = -21.98905 - 2.32492 * epsilon
        f_epsilon += 0.019650 * epsilon ** 2 + 13.76152 * np.sqrt(epsilon)

        return f_epsilon

    def set_seed(self, seed):
        """
        allow to set a new random seed
        """
        self._random_generator.seed(seed)

    def set_interpolation_factor(self, interp_factor):
        """
        set interpolation factor of charge-excess profiles
        """
        self._interp_factor = interp_factor

    def set_interpolation_factor2(self, interp_factor):
        """
        set interpolation factor around peak of form factor
        """
        self._interp_factor2 = interp_factor

    def get_time_trace(self, shower_energy, theta, N, dt, shower_type, n_index, R, shift_for_xmax=False,
                       same_shower=False, iN=None, output_mode='trace', theta_reference='X0'):
        """
        calculates the electric-field Askaryan pulse from a charge-excess profile

        Parameters
        ----------
        shower_energy: float
            the energy of the shower
        theta: float
            viewing angle, i.e., the angle between shower axis and launch angle of the signal (the ray path)
        N: int
            number of samples in the time domain
        dt: float
            size of one time bin in units of time
        profile_depth: array of floats
            shower depth values of the charge excess profile
        profile_ce: array of floats
            charge-excess values of the charge excess profile
        shower_type: string (default "HAD")
            type of shower, either "HAD" (hadronic), "EM" (electromagnetic) or "TAU" (tau lepton induced)
        n_index: float (default 1.78)
            index of refraction where the shower development takes place
        R: float (default 1km)
            observation distance, the signal amplitude will be scaled according to 1/R
        interp_factor: int (default 10)
            interpolation factor of charge-excess profile. Results in a more precise numerical integration which might be beneficial
            for small vertex distances but also slows down the calculation proportional to the interpolation factor.
        shift_for_xmax: bool (default True)
            if True the observer position is placed relative to the position of the shower maximum, if False it is placed
            with respect to (0,0,0) which is the start of the charge-excess profile
        same_shower: bool (default False)
            if False, for each request a new random shower realization is choosen.
            if True, the shower from the last request of the same shower type is used. This is needed to get the Askaryan
            signal for both ray tracing solutions from the same shower.
        iN: int or None (default None)
            specify shower number
        output_mode: string
            * 'trace' (default): return only the electric field trace
            * 'Xmax': return trace and position of xmax in units of length
            * 'full' return trace, depth and charge_excess profile
        theta_reference: string (default: X0)
            * 'X0': viewing angle relativ to start of the shower
            * 'Xmax': viewing angle is relativ to Xmax, internally it will be converted to be relative to X0

        Returns: array of floats
            array of electric-field time trace in 'on-sky' coordinate system eR, eTheta, ePhi
        """
        if not shower_type in self._library.keys():
            raise KeyError("shower type {} not present in library. Available shower types are {}".format(shower_type, *self._library.keys()))

        # determine closes available energy in shower library
        energies = np.array([*self._library[shower_type]])
        iE = np.argmin(np.abs(energies - shower_energy))
        rescaling_factor = shower_energy / energies[iE]
        logger.info("shower energy of {:.3g}eV requested, closest available energy is {:.3g}eV. The amplitude of the charge-excess profile will be rescaled accordingly by a factor of {:.2f}".format(shower_energy / units.eV, energies[iE] / units.eV, rescaling_factor))
        profiles = self._library[shower_type][energies[iE]]
        N_profiles = len(profiles['charge_excess'])

        if(iN is None):
            if(same_shower):
                if(shower_type in self._random_numbers):
                    iN = self._random_numbers[shower_type]
                    logger.info("using previously used shower {}/{}".format(iN, N_profiles))
                else:
                    logger.warning("no previous random number for shower type {} exists. Generating a new random number.".format(shower_type))
                    iN = self._random_generator.randint(N_profiles)
                    self._random_numbers[shower_type] = iN
                    logger.info("picking profile {}/{} randomly".format(iN, N_profiles))
            else:
                iN = self._random_generator.randint(N_profiles)
                self._random_numbers[shower_type] = iN
                logger.info("picking profile {}/{} randomly".format(iN, N_profiles))
        else:
            logger.info("using shower {}/{} as specified by user".format(iN, N_profiles))

        profile_depth = profiles['depth']
        profile_ce = profiles['charge_excess'][iN] * rescaling_factor

        xmax = profile_depth[np.argmax(profile_ce)]
        if(theta_reference == 'Xmax'):
            thetat = copy.copy(theta)
            theta = thetaprime_to_theta(theta, xmax, R)
            logger.info("transforming viewing angle from {:.2f} to {:.2f}".format(thetat / units.deg, theta / units.deg))
        elif(theta_reference != 'X0'):
            raise NotImplementedError("theta_reference = '{}' is not implemented".format(theta_reference))

        vp = self.get_vector_potential_fast(shower_energy, theta, N, dt, profile_depth, profile_ce, shower_type, n_index, R,
                                            self._interp_factor, self._interp_factor2, shift_for_xmax)
        trace = -np.diff(vp, axis=0) / dt
#         trace = -np.gradient(vp, axis=0) / dt

        # use viewing angle relative to shower maximum for rotation into spherical coordinate system (that reduced eR component)
        thetaprime = theta_to_thetaprime(theta, xmax, R)
        cs = cstrafo.cstrafo(zenith=thetaprime, azimuth=0)
        trace_onsky = cs.transform_from_ground_to_onsky(trace.T)
        if(output_mode == 'full'):
            return trace_onsky, profile_depth, profile_ce
        elif(output_mode == 'Xmax'):
            xmax = profile_depth[np.argmax(profile_ce)]
            Lmax = xmax / rho
            return trace_onsky, Lmax
        return trace_onsky

    def get_last_shower_profile_id(self):
        """
        returns dict
            the index of the randomly selected shower profile per shower type
            key is the shower type (string)
            value is the index (int)
        """
        return self._random_numbers

    def get_vector_potential_fast(self, shower_energy, theta, N, dt, profile_depth, profile_ce,
                                  shower_type="HAD", n_index=1.78, distance=1 * units.m,
                                  interp_factor=1., interp_factor2=100., shift_for_xmax=False):
        """
        fast interpolation of time-domain calculation of vector potential of the
        Askaryan pulse from a charge-excess profile

        Note that the returned array has N+1 samples so that the derivative (the efield) will have N samples.

        The numerical integration was replaces by a sum using the trapeoiz rule using vectorized numpy operations

        Parameters
        ----------
        shower_energy: float
            the energy of the shower
        theta: float
            viewing angle, i.e., the angle between shower axis and launch angle of the signal (the ray path)
        N: int
            number of samples in the time domain
        dt: float
            size of one time bin in units of time
        profile_depth: array of floats
            shower depth values of the charge excess profile
        profile_ce: array of floats
            charge-excess values of the charge excess profile
        shower_type: string (default "HAD")
            type of shower, either "HAD" (hadronic), "EM" (electromagnetic) or "TAU" (tau lepton induced)
        n_index: float (default 1.78)
            index of refraction where the shower development takes place
        distance: float (default 1km)
            observation distance, the signal amplitude will be scaled according to 1/R
        interp_factor: int (default 1)
            interpolation factor of charge-excess profile. Results in a more precise numerical integration which might be beneficial
            for small vertex distances but also slows down the calculation proportional to the interpolation factor.
            if None, the interpolation factor will be calculated from the distance
        interp_factor2: int (default 100)
            interpolation just around the peak of the form factor
        shift_for_xmax: bool (default True)
            if True the observer position is placed relative to the position of the shower maximum, if False it is placed
            with respect to (0,0,0) which is the start of the charge-excess profile
        """

        ttt = np.arange(0, (N + 1) * dt, dt)
        ttt = ttt + 0.5 * dt - ttt.mean()
        if(len(ttt) != N + 1):
            ttt = ttt[:-1]
        N = len(ttt)

        xn = n_index
        cher = np.arccos(1. / n_index)
        beta = 1.

        profile_dense = profile_depth
        profile_ce_interp = profile_ce
        if(interp_factor != 1):
            profile_dense = np.linspace(min(profile_depth), max(profile_depth), interp_factor * len(profile_depth))
            profile_ce_interp = np.interp(profile_dense, profile_depth, profile_ce)
        length = profile_dense / rho
        dxmax = length[np.argmax(profile_ce_interp)]
    #     theta2 = np.arctan(R * np.sin(theta)/(R * np.cos(theta) - dxmax))
    #     logger.warning("theta changes from {:.2f} to {:.2f}".format(theta/units.deg, theta2/units.deg))

        # calculate antenna position in ARZ reference frame
        # coordinate system is with respect to an origin which is located
        # at the position where the primary particle is injected in the medium. The reference frame
        # is z = along shower axis, and x,y are two arbitray directions perpendicular to z
        # and perpendicular among themselves of course.
        # For instance to place an observer at a distance R and angle theta w.r.t. shower axis in the x,z plane
        # it can be simply done by putting in the input file the numerical values:
        X = np.array([distance * np.sin(theta), 0., distance * np.cos(theta)])
        if(shift_for_xmax):
            logger.info("shower maximum at z = {:.1f}m, shifting observer position accordingly.".format(dxmax / units.m))
            X = np.array([distance * np.sin(theta), 0., distance * np.cos(theta) + dxmax])
        logger.info("setting observer position to {}".format(X))

        def get_dist_shower(X, z):
            """
            Distance from position in shower depth z' to each antenna.
            Denominator in Eq. (22) PRD paper

            Parameters
            ----------
            X: 3dim np. array
                position of antenna in ARZ reference frame
            z: shower depth
            """
            return (X[0] ** 2 + X[1] ** 2 + (X[2] - z) ** 2) ** 0.5

        # calculate total charged track length
        xntot = np.sum(profile_ce_interp) * (length[1] - length[0])
        factor = -xmu / (4. * np.pi)
        fc = 4. * np.pi / (xmu * np.sin(cher))

        vp = np.zeros((N, 3))
        for it, t in enumerate(ttt):
            tobs = t + (get_dist_shower(X, 0) / c * xn)
            z = length

            R = get_dist_shower(X, z)
            arg = z - (beta * c * tobs - xn * R)

            # Note that Acher peaks at tt=0 which corresponds to the observer time.
            # The shift from tobs to tt=0 is done when defining argument
            tt = (-arg / (c * beta))  # Parameterisation of A_Cherenkov with t in ns

            mask = abs(tt) < 20. * units.ns
            if(np.sum(mask) == 0):  #
                vp[it] = 0
                continue

            profile_dense2 = profile_dense
            profile_ce_interp2 = profile_ce_interp
            abc = False
            if(interp_factor2 != 1):
                # we only need to interpolate between +- 1ns to achieve a better precision in the numerical integration
                # the following code finds the indices sourrounding the bins fulfilling these condition
                # please not that we often have two distinct intervals having -1 < tt < 1
                tmask = (tt < 1 * units.ns) & (tt > -1 * units.ns)
                gaps = (tmask[1:] ^ tmask[:-1])  # xor
                indices = np.arange(len(gaps))[gaps]  # the indices in between tt is within -+ 1ns
                if(len(indices) != 0):  # only interpolate if we have time within +- 1 ns of the observer time
                    # now we add the corner cases of having the tt array start or end with an entry fulfilling the condition
                    if(len(indices) % 2 != 0):
                        if((tt[0] < 1 * units.ns) and (tt[0] > -1 * units.ns) and indices[0] != 0):
                            indices = np.append(0, indices)
                        else:
                            if(indices[-1] != (len(tt) - 1)):
                                indices = np.append(indices, len(tt) - 1)
                    if(len(indices) % 2 == 0):  # this rejects the cases where only the first or the last entry fulfills the -1 < tt < 1 condition
                        dt = tt[1] - tt[0]

                        dp = profile_dense2[1] - profile_dense2[0]
                        if(len(indices) == 2):  # we have only one interval
                            i_start = indices[0]
                            i_stop = indices[1]
                            profile_dense2 = np.arange(profile_dense[i_start], profile_dense[i_stop], dp / interp_factor2)
                            profile_ce_interp2 = np.interp(profile_dense2, profile_dense[i_start:i_stop], profile_ce_interp[i_start:i_stop])
                            profile_dense2 = np.append(np.append(profile_dense[:i_start], profile_dense2), profile_dense[i_stop:])
                            profile_ce_interp2 = np.append(np.append(profile_ce_interp[:i_start], profile_ce_interp2), profile_ce_interp[i_stop:])
                        elif(len(indices) == 4):  # we have two intervals, hence, we need to upsample two distinct intervals and put the full array back together.
                            i_start = indices[0]
                            i_stop = indices[1]
                            profile_dense2 = np.arange(profile_dense[i_start], profile_dense[i_stop], dp / interp_factor2)
                            profile_ce_interp2 = np.interp(profile_dense2, profile_dense[i_start:i_stop], profile_ce_interp[i_start:i_stop])

                            i_start3 = indices[2]
                            i_stop3 = indices[3]
                            profile_dense3 = np.arange(profile_dense[i_start3], profile_dense[i_stop3], dp / interp_factor2)
                            profile_ce_interp3 = np.interp(profile_dense3, profile_dense[i_start3:i_stop3], profile_ce_interp[i_start3:i_stop3])

                            profile_dense2 = np.append(np.append(np.append(np.append(
                                                            profile_dense[:i_start], profile_dense2),
                                                               profile_dense[i_stop:i_start3]),
                                                                  profile_dense3),
                                                                       profile_dense[i_stop3:])
                            profile_ce_interp2 = np.append(np.append(np.append(np.append(
                                                    profile_ce_interp[:i_start],
                                                    profile_ce_interp2),
                                                    profile_ce_interp[i_stop:i_start3]),
                                                    profile_ce_interp3),
                                                    profile_ce_interp[i_stop3:])

                        else:
                            raise NotImplementedError("length of indices is not 2 nor 4")  # this should never happen
                        if 0:
                            abc = True
                            i_stop = len(profile_dense) - 1
                            from matplotlib import pyplot as plt
                            fig, ax = plt.subplots(1, 1)
                            ax.plot(tt, color='0.5')
                            ax.plot(np.arange(len(tmask))[tmask], tt[tmask], 'o')
                            ax.plot(indices, np.ones_like(indices), 'd')
            #                 ax.plot(np.arange(len(tmask))[gaps], tt[gaps], 'd')
                            plt.show()

                        # recalculate parameters for interpolated values
                        z = profile_dense2 / rho
                        R = get_dist_shower(X, z)
                        arg = z - (beta * c * tobs - xn * R)
                        tt = (-arg / (c * beta))
                        mask = abs(tt) < 20. * units.ns
                        tmask = (tt < 1 * units.ns) & (tt > -1 * units.ns)

            F_p = np.zeros_like(tt)
            # Cut fit above +/-5 ns

            u_x = X[0] / R
            u_y = X[1] / R
            u_z = (X[2] - z) / R
            beta_z = 1.
            vperp_x = u_x * u_z * beta_z
            vperp_y = u_y * u_z * beta_z
            vperp_z = -(u_x * u_x + u_y * u_y) * beta_z
            v = np.array([vperp_x, vperp_y, vperp_z])
            """
            Function F_p Eq.(15) PRD paper.
            """
            # Factor accompanying the F_p in Eq.(15) in PRD paper
            beta = 1.
            if(np.sum(mask)):
                # Choose Acher between purely electromagnetic, purely hadronic or mixed shower
                # Eq.(16) PRD paper.
                E_TeV = shower_energy / units.TeV
                Acher = np.zeros_like(tt)
                if(shower_type == "HAD"):
                    mask2 = tt > 0 & mask
                    if(np.sum(mask2)):
                        Acher[mask2] = self._Af_p * E_TeV * (np.exp(-np.abs(tt[mask2]) / self._t0_p_pos) +
                                              (1. + self._freq_p_pos * np.abs(tt[mask2])) ** self._exp_p_pos)  # hadronic
                    mask2 = tt <= 0 & mask
                    if(np.sum(mask2)):
                        Acher[mask2] = self._Af_p * E_TeV * (np.exp(-np.abs(tt[mask2]) / self._t0_p_neg) +
                                              (1. + self._freq_p_neg * np.abs(tt[mask2])) ** self._exp_p_neg)  # hadronic
                    Acher *= self.em_fraction(shower_energy)
                elif(shower_type == "EM"):
                    mask2 = tt > 0 & mask
                    if(np.sum(mask2)):
                        Acher[mask2] = self._Af_e * E_TeV * (np.exp(-np.abs(tt[mask2]) / self._t0_e_pos) +
                                              (1. + self._freq_e_pos * np.abs(tt[mask2])) ** self._exp_e_pos)  # electromagnetic
                    mask2 = tt <= 0 & mask
                    if(np.sum(mask2)):
                        Acher[mask2] = self._Af_e * E_TeV * (np.exp(-np.abs(tt[mask2]) / self._t0_e_neg) +
                                              (1. + self._freq_e_neg * np.abs(tt[mask2])) ** self._exp_e_neg)  # electromagnetic
                elif(shower_type == "TAU"):
                    logger.error("Tau showers are not yet implemented")
                    raise NotImplementedError("Tau showers are not yet implemented")
                else:
                    msg = "showers of type {} are not implemented. Use 'HAD', 'EM' or 'TAU'".format(shower_type)
                    logger.error(msg)
                    raise NotImplementedError(msg)
                # Obtain "shape" of Lambda-function from vp at Cherenkov angle
                # xntot = LQ_tot in PRD paper
                F_p[mask] = Acher[mask] * fc / xntot
    #         F_p[~mask] = 1.e-30 * fc / xntot
            F_p[~mask] = 0

            vp[it] = np.trapz(-v * profile_ce_interp2 * F_p / R, z)
            if  0:
                import matplotlib.pyplot as plt
                fig, ax = plt.subplots(1, 1)
                inte = -v * profile_ce_interp2 * F_p / R
                ax.plot(tt, inte[0], '-')
                ax.plot(tt, inte[1], '-')
                ax.plot(tt, inte[2], '-')
                ax.plot(tt[tmask], inte[0][tmask], 'o')
                ax.plot(tt[tmask], inte[1][tmask], 'o')
                ax.plot(tt[tmask], inte[2][tmask], 'o')
                ax.set_title("{}".format(vp[it]))
                plt.show()

        vp *= factor
        if 0:
            import matplotlib.pyplot as plt
            fig, (ax, ax2) = plt.subplots(1, 2)
            ax.plot(vp)
            print(vp.shape)
            t0 = -np.gradient(vp.T[0]) / dt
            t1 = -np.gradient(vp.T[1]) / dt
            t2 = -np.gradient(vp.T[2]) / dt
            trace2 = -np.diff(vp, axis=0) / dt
    #         print(trace.shape)
            ax2.plot(t0)
            ax2.plot(t1)
            ax2.plot(t2)

            ax2.plot(trace2.T[0], '--')
            ax2.plot(trace2.T[1], '--')
            ax2.plot(trace2.T[2], '--')
            plt.show()
        return vp

    def get_vector_potential(self, energy, theta, N, dt, y=1, ccnc='cc', flavor=12, n_index=1.78, R=1 * units.m,
                             profile_depth=None, profile_ce=None):
        """
        python transcription of original FORTRAN code
        """

        tt = np.arange(0, (N + 1) * dt, dt)
        tt = tt + 0.5 * dt - tt.mean()
    #     tmin = tt.min()
    #     tmax = tt.max()

    #     tmin = -100 * units.ns
    #     tmax = 100 * units.ns

    #     tt = np.arange(tmin, tmax, dt)
    #     tt += 0.5 * dt
        N = len(tt)

        xn = n_index
        cher = np.arccos(1. / n_index)
        beta = 1.

        # calculate antenna position in ARZ reference frame
        # coordinate system is with respect to an origin which is located
        # at the position where the primary particle is injected in the medium. The reference frame
        # is z = along shower axis, and x,y are two arbitray directions perpendicular to z
        # and perpendicular among themselves of course.
        # For instance to place an observer at a distance R and angle theta w.r.t. shower axis in the x,z plane
        # it can be simply done by putting in the input file the numerical values:
        X = np.array([R * np.sin(theta), 0., R * np.cos(theta)])

        def get_dist_shower(X, z):
            """
            Distance from position in shower depth z' to each antenna.
            Denominator in Eq. (22) PRD paper

            Parameters
            ----------
            X: 3dim np. array
                position of antenna in ARZ reference frame
            z: shower depth
            """
            return (X[0] ** 2 + X[1] ** 2 + (X[2] - z) ** 2) ** 0.5

        length = profile_depth / rho
        xnep = intp.interp1d(length, profile_ce, bounds_error=False, fill_value=0)

        # calculate total charged track length
        xntot = np.sum(profile_ce) * (length[1] - length[0])
        # print("{:.5g}".format(xntot))
        # res = int.quad(xnep, length.min(), length.max())
        # print("{:.5g} {:.5g}".format(*res))

        if 0:  # debug plot
            ll = np.linspace(length.min(), length.max(), 10000)
            plt.plot(ll, xnep(ll))
            plt.plot(length, N_e - N_p, 'o')
            plt.show()

        factor = -xmu / (4. * np.pi)

        def xintegrand(z, index, tobs):
            R = get_dist_shower(X, z)
            arg = z - (beta * c * tobs - xn * R)
            u_x = X[0] / R
            u_y = X[1] / R
            u_z = (X[2] - z) / R
            beta_z = 1.
            vperp_x = u_x * u_z * beta_z
            vperp_y = u_y * u_z * beta_z
            vperp_z = -(u_x * u_x + u_y * u_y) * beta_z
            v = np.array([vperp_x, vperp_y, vperp_z])[index]

            return -v * xnep(z) * F_p(arg) / R

        def F_p(arg):
            """
            Function F_p Eq.(15) PRD paper.
            """
            # Factor accompanying the F_p in Eq.(15) in PRD paper
            fc = 4. * np.pi / (xmu * np.sin(cher))
            beta = 1.

            # Note that Acher peaks at tt=0 which corresponds to the observer time.
            # The shift from tobs to tt=0 is done when defining argument
            tt = (-arg / (c * beta))  # Parameterisation of A_Cherenkov with t in ns
            # Cut fit above +/-5 ns
            if (abs(tt) > 5. * units.ns):
                return 1.e-30 * fc / xntot

            # Choose Acher between purely electromagnetic, purely hadronic or mixed shower
            # Eq.(16) PRD paper.
            E_TeV = energy / units.TeV
            em_fraction = self.em_fraction(energy)
            if (tt > 0):
                A_e = self._Af_e * E_TeV * (np.exp(-np.abs(tt) / self._t0_e_pos) +
                                      (1. + self._freq_e_pos * np.abs(tt)) ** self._exp_e_pos)  # electromagnetic
                A_p = self._Af_p * E_TeV * em_fraction * (np.exp(-np.abs(tt) / self._t0_p_pos) +
                                      (1. + self._freq_p_pos * np.abs(tt)) ** self._exp_p_pos)  # hadronic
            else:
                A_e = self._Af_e * E_TeV * (np.exp(-np.abs(tt) / self._t0_e_neg) +
                                      (1. + self._freq_e_neg * np.abs(tt)) ** self._exp_e_neg)  # electromagnetic
                A_p = self._Af_p * E_TeV * em_fraction * (np.exp(-np.abs(tt) / self._t0_p_pos) +
                                      (1. + self._freq_p_neg * np.abs(tt)) ** self._exp_p_neg)  # hadronic

            if(ccnc == 'nc'):
                Acher = y * A_p
            else:
                if(np.abs(flavor) == 12):
                    Acher = (1. - y) * A_e + y * A_p
                else:
                    Acher = 0

            # Obtain "shape" of Lambda-function from vp at Cherenkov angle
            # xntot = LQ_tot in PRD paper
            return Acher * fc / xntot

        vp = np.zeros((N, 3))
        for it, t in enumerate(tt):
            tobs = t + (get_dist_shower(X, 0) / c * xn)
            xmin = length.min()
            xmax = length.max()
            if(X[0] != 0):
                vp[it][0] = int.quad(xintegrand, xmin, xmax, args=(0, tobs))[0]
            if(X[1] != 0):
                vp[it][1] = int.quad(xintegrand, xmin, xmax, args=(1, tobs))[0]
            if(X[2] != 0):
                vp[it][2] = int.quad(xintegrand, xmin, xmax, args=(2, tobs))[0]
        vp *= factor
        return vp


class ARZ_tabulated(object):
    __instance = None

    def __new__(cls, seed=1234, library=None):
        if ARZ_tabulated.__instance is None:
            ARZ_tabulated.__instance = object.__new__(cls, seed, library)
        return ARZ_tabulated.__instance

    def __init__(self, seed=1234, library=None):
        logger.warning("setting seed to {}".format(seed))
        self._random_generator = np.random.RandomState(seed)
        self._random_numbers = {}
        self._version = (1, 1)
        # # load shower library into memory
        if(library is None):
            library = os.path.join(os.path.dirname(__file__), "shower_library/ARZ_library_v{:d}.{:d}.pkl".format(*self._version))
        else:
            if(not os.path.exists(library)):
                logger.error("user specified pulse library {} not found.".format(library))
                raise FileNotFoundError("user specified pulse library {} not found.".format(library))
        self.__check_and_get_library()

        logger.warning("loading pulse library into memory")
        self._library = io_utilities.read_pickle(library)

    def __check_and_get_library(self):
        """
        checks if pulse library exists and is up to date by comparing the sha1sum. If the library does not exist
        or changes on the server, a new library will be downloaded.
        """
        path = os.path.join(os.path.dirname(__file__), "shower_library/ARZ_library_v{:d}.{:d}.pkl".format(*self._version))

        download_file = False
        if(not os.path.exists(path)):
            logger.warning("ARZ library version {} does not exist on the local file system yet. It will be downloaded to {}".format(self._version, path))
            download_file = True

        if(os.path.exists(path)):
            BUF_SIZE = 65536 * 2 ** 4  # lets read stuff in 64kb chunks!
            import hashlib
            import json
            sha1 = hashlib.sha1()
            with open(path, 'rb') as f:
                while True:
                    data = f.read(BUF_SIZE)
                    if not data:
                        break
                    sha1.update(data)

            shower_directory = os.path.join(os.path.dirname(__file__), "shower_library/")
            with open(os.path.join(shower_directory, 'shower_lib_hash.json'), 'r') as fin:
                lib_hashs = json.load(fin)
                if("ARZ_{:d}.{:d}".format(*self._version) in lib_hashs.keys()):
                    if(sha1.hexdigest() != lib_hashs["{:d}.{:d}".format(*self._version)]):
                        logger.warning("pulse library {} has changed on the server. downloading newest version...".format(self._version))
                        download_file = True
                else:
                    logger.warning("no hash sum of {} available, skipping up-to-date check".format(os.path.basename(path)))
        if not download_file:
            return True
        else:
            import requests
            URL = 'http://arianna.ps.uci.edu/~arianna/data/ce_shower_library/ARZ_library_v{:d}.{:d}.pkl'.format(*self._version)

            logger.info("downloading pulse library {} from {}. This can take a while...".format(self._version, URL))
            r = requests.get(URL)
            if (r.status_code != requests.codes.ok):
                logger.error("error in download of antenna model")
                raise IOError("error in download of antenna model")
            with open(path, "wb") as code:
                code.write(r.content)
            logger.info("...download finished.")

    def set_seed(self, seed):
        """
        allow to set a new random seed
        """
        self._random_generator.seed(seed)

    def get_time_trace(self, shower_energy, theta, N, dt, shower_type, n_index, R,
                       same_shower=False, iN=None, output_mode='trace', theta_reference='X0'):
        """
        calculates the electric-field Askaryan pulse from a charge-excess profile

        Parameters
        ----------
        shower_energy: float
            the energy of the shower
        theta: float
            viewing angle, i.e., the angle between shower axis and launch angle of the signal (the ray path)
        N: int
            number of samples in the time domain
        dt: float
            size of one time bin in units of time
        shower_type: string (default "HAD")
            type of shower, either "HAD" (hadronic), "EM" (electromagnetic) or "TAU" (tau lepton induced)
        n_index: float (default 1.78)
            index of refraction where the shower development takes place
        R: float (default 1km)
            observation distance, the signal amplitude will be scaled according to 1/R
        same_shower: bool (default False)
            if False, for each request a new random shower realization is choosen.
            if True, the shower from the last request of the same shower type is used. This is needed to get the Askaryan
            signal for both ray tracing solutions from the same shower.
        iN: int or None (default None)
            specify shower number
        output_mode: string
            * 'trace' (default): return only the electric field trace
            * 'Xmax': return trace and position of xmax in units of length
        theta_reference: string (default: X0)
            * 'X0': viewing angle relativ to start of the shower
            * 'Xmax': viewing angle is relativ to Xmax, internally it will be converted to be relative to X0

        Returns: array of floats
            array of electric-field time trace in 'on-sky' coordinate system eR, eTheta, ePhi
        """
        if not shower_type in self._library.keys():
            raise KeyError("shower type {} not present in library. Available shower types are {}".format(shower_type, *self._library.keys()))

        # determine closes available energy in shower library
        energies = np.array(list(self._library[shower_type].keys()))
        iE = np.argmin(np.abs(energies - shower_energy))
        rescaling_factor = shower_energy / energies[iE]
        logger.info("shower energy of {:.3g}eV requested, closest available energy is {:.3g}eV. The pulse amplitude will be rescaled accordingly by a factor of {:.2f}".format(shower_energy / units.eV, energies[iE] / units.eV, rescaling_factor))
        profiles = self._library[shower_type][energies[iE]]
        N_profiles = len(profiles.keys())

        if(iN is None):
            if(same_shower):
                if(shower_type in self._random_numbers):
                    iN = self._random_numbers[shower_type]
                    logger.info("using previously used shower {}/{}".format(iN, N_profiles))
                else:
                    logger.warning("no previous random number for shower type {} exists. Generating a new random number.".format(shower_type))
                    iN = self._random_generator.randint(N_profiles)
                    self._random_numbers[shower_type] = iN
                    logger.info("picking profile {}/{} randomly".format(iN, N_profiles))
            else:
                iN = self._random_generator.randint(N_profiles)
                self._random_numbers[shower_type] = iN
                logger.info("picking profile {}/{} randomly".format(iN, N_profiles))
        else:
            logger.info("using shower {}/{} as specified by user".format(iN, N_profiles))

        thetas = profiles[iN].keys()
        iT = np.argmin(np.abs(thetas - theta))
        logger.info("selecting theta = {:.2f} ({:.2f} requested)".format(thetas[iT] / units.deg, theta))
        trace = profiles[iT]['trace']
        t0 = profiles[iT]['t0']
        Lmax = profiles[iT]['Lmax']
        trace2 = np.zeros(N)
        tcenter = N // 2 * dt
        tstart = t0 + tcenter
        i0 = np.int(np.round(tstart / dt))
        trace2[i0:(i0 + len(trace))] = trace

        trace2 *= self._library['meta']['R'] / R * rescaling_factor

        if(output_mode == 'Xmax'):
            return trace2, Lmax
        return trace2



from __future__ import print_function, division
import numpy as np
from NuRadioReco.utilities import units
from NuRadioReco.utilities import io_utilities
import os
from scipy import interpolate as intp
import glob
import pickle
import sys
from radiotools import plthelpers as php
from matplotlib import pyplot as plt
from NuRadioMC.SignalGen.ARZ import ARZ
import logging
from scipy import signal
import matplotlib.gridspec as gridspec

logger = logging.getLogger("test")
logging.basicConfig(level=logging.WARNING)
rho = 0.924 * units.g / units.cm ** 3  # density g cm^-3

plot=False

if __name__ == "__main__":
    sampling_rate = 5 * units.GHz
    dt = 1. / sampling_rate
    T2 = 50 * units.ns
    N2 = np.int(np.round(T2 / dt))
    n_index = 1.78
    theta = np.arccos(1. / n_index)
    R = 5 * units.km
    T = 600 * units.ns
    N = np.int(np.round(T / dt))
    tt = np.arange(0, dt * N, dt)

    dCs = np.append(np.arange(-20, -5, 1), np.append(np.arange(-5,-1,0.2), np.arange(-1, 0, .05)))
    dCs = np.append(np.append(dCs, [0]), -1 * dCs)
    dCs = np.sort(dCs) * units.deg
    print('generating Askaryan pulses for the following viewing angles')
    print(dCs/units.deg)

    lib_path = "/Users/cglaser/work/ARIANNA/data/ARZ/v1.1/library_v1.1.pkl"
    a = ARZ.ARZ(library=lib_path, interp_factor=100)

    showers = {}
    showers['meta'] = {'dt': dt,
                       'n_index': n_index,
                       'R': R}

    lib = io_utilities.read_pickle(lib_path)
    for iS, shower_type in enumerate(lib):  # loop through shower types
        if(shower_type not in showers):
            showers[shower_type] = {}
        print("shower type {}".format(shower_type))
        b = []
        nb = []
        for iE, E in enumerate(lib[shower_type]):  # loop through energies
            print('E = {:.2g}eV'.format(E))

            nE = len(lib[shower_type][E]['charge_excess'])
            if(E not in showers[shower_type]):
                showers[shower_type][E] = {}
            for iN in range(nE):  # loop through shower realizations
                if(iN not in showers[shower_type][E]):
                    showers[shower_type][E][iN] = {}
                if plot:
                    fig, ax = plt.subplots(1, 1)
                    fig2, ax2 = plt.subplots(1, 1)
                for dC in dCs:  # loop through different cherenkov angles
                    theta1 = theta + dC
                    trace, Lmax = a.get_time_trace(E, theta1, N, dt, shower_type, n_index, R, iN=iN, output_mode='Xmax')
                    iMax = np.argmax(np.abs(trace[1]))
                    i1 = iMax - N2 // 2
                    if((i1 + N2) >= N) or (i1 < 0):
                        print("tracelength to short, i1 = {}".format(i1))
#                             raise IndexError("tracelength to short, i1 = {}".format(i1))
                    t0 = tt[i1] - tt.mean()
                    showers[shower_type][E][iN][theta1] = {}
                    showers[shower_type][E][iN][theta1]['t0'] = t0
                    showers[shower_type][E][iN][theta1]['Lmax'] = Lmax
                    showers[shower_type][E][iN][theta1]['trace'] = trace[1][i1:(i1 + N2)]
                    if(plot):
                        ax.plot(tt[i1:(i1 + N2)], trace[1][i1:(i1 + N2)] / np.abs(trace[1]).max())
                        ax2.plot(tt[i1:(i1 + N2)], np.abs(signal.hilbert(trace[1][i1:(i1 + N2)])))
                if(plot):
                    ax.set_title("{} E = {:.2g}eV i = {:d}".format(shower_type, E / units.eV, iN))
                    ax2.set_title("{} E = {:.2g}eV i = {:d}".format(shower_type, E / units.eV, iN))
                    ax.set_xlabel('time [ns]')
                    ax2.set_xlabel('time [ns]')
                    ax.set_ylabel('normalized amplitude')
                    ax2.semilogy(True)
                    fig.tight_layout()
                    fig2.tight_layout()
#                         plt.show()
                    fig.savefig("plots/{}_{:.2g}eV_{:03d}.png".format(shower_type, E / units.eV, iN))
                    fig2.savefig("plots/{}_{:.2g}eV_{:03d}_log.png".format(shower_type, E / units.eV, iN))
                    plt.close("all")

        with open("ARZ_library_v1.1.pkl", 'wb') as fout:
            pickle.dump(showers, fout, protocol=2)



#!/usr/bin/env python
import sys
from numpy import testing
import numpy as np

import NuRadioReco.modules.io.eventReader
from NuRadioReco.utilities import units

try:
    file1 = sys.argv[1]
    file2 = sys.argv[2]
except:
    print("No files given")
    sys.exit(-1)

print("Testing the files {} and {} for equality".format(file1, file2))

def all_traces(file):
    eventReader1 = NuRadioReco.modules.io.eventReader.eventReader()
    eventReader1.begin(file)
    i = 0
    for iE1, event1 in enumerate(eventReader1.run()):
        for st1, station1 in enumerate(event1.get_stations()):
            for channel1 in station1.iter_channels():
                trace1 = channel1.get_trace()
                if i == 0:
                    all_traces = trace1
                else:
                    all_traces.append(trace1)
    return all_traces

all_traces_1 = all_traces(file1)
all_traces_2 = all_traces(file2)

diff = all_traces_1 - all_traces_2

if np.any(diff != 0):
    print("The arrays are different, difference in traces:", diff)

print("Maximum difference between traces [mV]", np.max(np.abs(diff))/units.mV)

precision = 7
testing.assert_almost_equal(all_traces_1, all_traces_2,decimal=precision)

try:
    testing.assert_equal(all_traces_1, all_traces_2)
except:
    print("Traces agree within {} decimals, but not completely identical".format(precision))

print("Traces are identical")




from __future__ import absolute_import, division, print_function
import argparse
from six import iteritems
# import detector simulation modules
import NuRadioReco.modules.efieldToVoltageConverter
import NuRadioReco.modules.trigger.simpleThreshold
import NuRadioReco.modules.channelResampler
import NuRadioReco.modules.channelBandPassFilter
import NuRadioReco.modules.channelGenericNoiseAdder

from NuRadioReco.utilities import units
from NuRadioMC.simulation import simulation
from NuRadioReco.framework.parameters import electricFieldParameters as efp
import logging
import numpy as np
logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger("runDeltaTStudy")

# initialize detector sim modules
efieldToVoltageConverter = NuRadioReco.modules.efieldToVoltageConverter.efieldToVoltageConverter()
efieldToVoltageConverter.begin(pre_pulse_time=0 * units.ns, post_pulse_time=0 * units.ns)
triggerSimulator = NuRadioReco.modules.trigger.simpleThreshold.triggerSimulator()
channelResampler = NuRadioReco.modules.channelResampler.channelResampler()
channelBandPassFilter = NuRadioReco.modules.channelBandPassFilter.channelBandPassFilter()
channelGenericNoiseAdder = NuRadioReco.modules.channelGenericNoiseAdder.channelGenericNoiseAdder()


class mySimulation(simulation.simulation):

    def _detector_simulation(self):

        # start detector simulation
        efieldToVoltageConverter.run(self._evt, self._station, self._det)  # convolve efield with antenna pattern

        # downsample trace to internal simulation sampling rate (the efieldToVoltageConverter upsamples the trace to
        # 20 GHz by default to achive a good time resolution when the two signals from the two signal paths are added)
        channelResampler.run(self._evt, self._station, self._det, sampling_rate=1. / self._dt)

        if self._is_simulate_noise():
            max_freq = 0.5 / self._dt
            norm = self._get_noise_normalization(self._station.get_id())  # assuming the same noise level for all stations
            Vrms = self._Vrms / (norm / (max_freq)) ** 0.5  # normalize noise level to the bandwidth its generated for
            channelGenericNoiseAdder.run(self._evt, self._station, self._det, amplitude=Vrms, min_freq=0 * units.MHz,
                                         max_freq=max_freq, type='rayleigh')

        # bandpass filter trace, the upper bound is higher then the sampling rate which makes it just a highpass filter
        channelBandPassFilter.run(self._evt, self._station, self._det, passband=[80 * units.MHz, 1000 * units.GHz],
                                  filter_type='butter', order=2)
        channelBandPassFilter.run(self._evt, self._station, self._det, passband=[0, 500 * units.MHz],
                                  filter_type='butter', order=10)
        triggerSimulator.run(self._evt, self._station, self._det,
                             threshold=2 * self._Vrms,
                             triggered_channels=None,
                             number_concidences=1,
                             trigger_name='pre_trigger_2sigma')

        # downsample trace back to detector sampling rate
        channelResampler.run(self._evt, self._station, self._det, sampling_rate=self._sampling_rate_detector)


parser = argparse.ArgumentParser(description='Run NuRadioMC simulation')
parser.add_argument('inputfilename', type=str,
                    help='path to NuRadioMC input event list')
parser.add_argument('detectordescription', type=str,
                    help='path to file containing the detector description')
parser.add_argument('config', type=str,
                    help='NuRadioMC yaml config file')
parser.add_argument('outputfilename', type=str,
                    help='hdf5 output filename')
parser.add_argument('outputfilenameNuRadioReco', type=str, nargs='?', default=None,
                    help='outputfilename of NuRadioReco detector sim file')
args = parser.parse_args()

sim = mySimulation(inputfilename=args.inputfilename,
                            outputfilename=args.outputfilename,
                            detectorfile=args.detectordescription,
                            outputfilenameNuRadioReco=args.outputfilenameNuRadioReco,
                            config_file=args.config,
                            file_overwrite=True)
sim.run()



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function
from NuRadioReco.utilities import units
from NuRadioMC.EvtGen.generator import generate_eventlist_cylinder

# define simulation volume
zmin = -2.7 * units.km  # the ice sheet at South Pole is 2.7km deep
zmax = 0 * units.km
rmin = 0 * units.km
rmax = 2 * units.km

# generate one event list at 1e18 eV with 10000 neutrinos
generate_eventlist_cylinder('1e18_full.hdf5', 2000, 1e18 * units.eV, 1e18 * units.eV, rmin, rmax, zmin, zmax,
                            full_rmin=rmin, full_rmax=rmax, full_zmin=zmin, full_zmax=zmax)



from __future__ import absolute_import, division, print_function
import numpy as np
from radiotools import helper as hp
from radiotools import plthelpers as php
from matplotlib import pyplot as plt
from NuRadioReco.utilities import units
import h5py
import argparse
import json
import time
import os

parser = argparse.ArgumentParser(description='Plot NuRadioMC event list input')
parser.add_argument('inputfilename', type=str, nargs='+',
                    help='path to NuRadioMC hdf5 simulation input')
args = parser.parse_args()

# data for vertex plot
xx = []
yy = []
zz = []
zeniths = []
azimuths = []
inelasticity = []
flavors = []
interaction_type = []

filename = os.path.splitext(os.path.basename(args.inputfilename[0]))[0]
dirname = os.path.dirname(args.inputfilename[0])
plot_folder = os.path.join(dirname, 'plots', 'input', filename)
if(not os.path.exists(plot_folder)):
    os.makedirs(plot_folder)

for input_filename in args.inputfilename:
    print(f"parsing file {input_filename}")
    fin = h5py.File(input_filename, 'r')
    xx.extend(np.array(fin['xx']))
    yy.extend(np.array(fin['yy']))
    zz.extend(np.array(fin['zz']))
    zeniths.extend(np.array(fin['zeniths']))
    azimuths.extend(np.array(fin['azimuths']))
    inelasticity.extend(np.array(fin['inelasticity']))
    flavors.extend(np.array(fin['flavors']))
    interaction_type.extend(np.array(fin['interaction_type']))

print(f"starting plotting")
###########################
# plot flavor ratios
###########################
flavor_labels = ['e cc', r'$\bar{e}$ cc', 'e nc', r'$\bar{e}$ nc',
           '$\mu$ cc', r'$\bar{\mu}$ cc', '$\mu$ nc', r'$\bar{\mu}$ nc',
           r'$\tau$ cc', r'$\bar{\tau}$ cc', r'$\tau$ nc', r'$\bar{\tau}$ nc']
flavors = np.array(flavors)
interaction_type = np.array(interaction_type)
flavor_sum = np.zeros(len(flavor_labels))
flavor_sum[0] = np.sum((flavors == 12) & (interaction_type == b'cc'))
flavor_sum[1] = np.sum((flavors == -12) & (interaction_type == b'cc'))
flavor_sum[2] = np.sum((flavors == 12) & (interaction_type == b'nc'))
flavor_sum[3] = np.sum((flavors == -12) & (interaction_type == b'nc'))

flavor_sum[4] = np.sum((flavors == 14) & (interaction_type == b'cc'))
flavor_sum[5] = np.sum((flavors == -14) & (interaction_type == b'cc'))
flavor_sum[6] = np.sum((flavors == 14) & (interaction_type == b'nc'))
flavor_sum[7] = np.sum((flavors == -14) & (interaction_type == b'nc'))

flavor_sum[8] = np.sum((flavors == 16) & (interaction_type == b'cc'))
flavor_sum[9] = np.sum((flavors == -16) & (interaction_type == b'cc'))
flavor_sum[10] = np.sum((flavors == 16) & (interaction_type == b'nc'))
flavor_sum[11] = np.sum((flavors == -16) & (interaction_type == b'nc'))

fig, ax = plt.subplots(1, 1, figsize=(8, 6))
ax.bar(range(len(flavor_labels)), flavor_sum)
ax.set_xticks(range(len(flavor_labels)))
ax.set_xticklabels(flavor_labels, fontsize='large', rotation=45)
ax.set_ylabel('weighted number of triggers', fontsize='large')
fig.tight_layout()
fig.savefig(os.path.join(plot_folder, 'flavor.png'))

# plot vertex distribution
fig, ax = plt.subplots(1, 1)
xx = np.array(xx)
yy = np.array(yy)
rr = (xx ** 2 + yy ** 2) ** 0.5
zz = np.array(zz)
h = ax.hist2d(rr / units.m, zz / units.m, bins=[np.arange(0, 4000, 100), np.arange(-3000, 0, 100)],
              cmap=plt.get_cmap('Blues'))
cb = plt.colorbar(h[3], ax=ax)
cb.set_label("number of events")
ax.set_aspect('equal')
ax.set_xlabel("r [m]")
ax.set_ylabel("z [m]")
fig.tight_layout()
plt.title('vertex distribution')
plt.savefig(os.path.join(plot_folder, "simInputVertex.png"))

# plot incoming direction
zeniths = np.array(zeniths)
azimuths = np.array(azimuths)
fig, axs = php.get_histograms([zeniths / units.deg, azimuths / units.deg],
                              bins=[np.arange(0, 181, 2), np.arange(0, 361, 5)],
                              xlabels=['zenith [deg]', 'azimuth [deg]'],
                              stats=False)
fig.suptitle('neutrino direction')
fig.subplots_adjust(top=0.9)

plt.savefig(os.path.join(plot_folder, "simInputIncoming.png"))

fig, ax = php.get_histogram(np.cos(zeniths), bins=np.arange(-1, 1.01, 0.1),
                            xlabel="cos zenith")
plt.savefig(os.path.join(plot_folder, "simInputIncoming_coszenith.png"))
# plot inelasticity
inelasticity = np.array(inelasticity)
fig, axs = php.get_histogram(inelasticity,
                             bins=np.logspace(np.log10(0.0001), np.log10(1.0), 50),
                             xlabel='inelasticity', figsize=(6, 6),
                             stats=True)
axs.semilogx(True)
plt.title('inelasticity')
plt.savefig(os.path.join(plot_folder, "simInputInelasticity.png"))



import pyPROPOSAL as pp
import numpy as np
from NuRadioReco.utilities import units
import os

"""
This module takes care of the PROPOSAL implementation. Some important things
should be considered.
Units: PROPOSAL used a fixed system of units that differs from that of NuRadioMC
and conversion between them must be done carefully. The definition of PROPOSAL
units can be found in this file. The most important are the energy unit (MeV)
and the distance unit (cm).
When a muon or a tau is propagated using PROPOSAL and its secondaries are obtained,
most of the secondaries belong to a DynamicData type. The ones we should care
about are the following:
- Brems: a bremsstrahlung photon
- DeltaE: an ionized electron
- EPair: an electron/positron pair
- Hadrons: a set of unspecified hadrons
- NuclInt: the products of a nuclear interaction
The last secondaries obtained via the propagation belong to the Particle DynamicData
type, and represent the products of the decay. They are standard particles with
a PDG code.
"""

# Units definition in PROPOSAL
pp_eV  = 1.e-6
pp_keV = 1.e-3
pp_MeV = 1.e0
pp_GeV = 1.e3
pp_TeV = 1.e6
pp_PeV = 1.e9
pp_EeV = 1.e12
pp_ZeV = 1.e15

pp_m = 1.e2
pp_km = 1.e5

class SecondaryProperties:
    """
    This class stores the properties from secondary particles that are
    relevant for NuRadioMC, namely:
    - distance, the distance to the first interaction vertex
    - energy, the particle energy
    - shower_type, whether the shower they induce is hadronic or electromagnetic
    - name, its name according to the particle_name dictionary on this module

    Distance and energy are expected to be in NuRadioMC units
    """
    def __init__(self,
                 distance,
                 energy,
                 shower_type,
                 code,
                 name):
        self.distance = distance
        self.energy = energy
        self.shower_type = shower_type
        self.code = code
        self.name = name

    def __str__(self):
        s  = "Particle and code: {:} ({:})\n".format(self.name, self.code)
        s += "Energy in PeV: {:}\n".format(self.energy/units.PeV)
        s += "Distance from vertex in km: {:}\n".format(self.distance/units.km)
        s += "Shower type: {:}\n".format(self.shower_type)
        return s

def particle_code (particle):
    """
    If a particle object from PROPOSAL is passed as input, it returns the
    corresponding PDG particle code. DynamicData objects are not considered
    particles internally in PROPOSAL, so they are handled differently.

    Parameters
    ----------
    particle: particle object from PROPOSAL

    Returns
    -------
    integer with the particle code. None if the argument is not a particle
    """
    if particle.particle_def == pp.particle.GammaDef.get()      : return    0
    elif particle.particle_def == pp.particle.EMinusDef.get()   : return   11
    elif particle.particle_def == pp.particle.EPlusDef.get()    : return  -11
    elif particle.particle_def == pp.particle.NuEDef.get()      : return   12
    elif particle.particle_def == pp.particle.NuEBarDef.get()   : return  -12
    elif particle.particle_def == pp.particle.MuMinusDef.get()  : return   13
    elif particle.particle_def == pp.particle.MuPlusDef.get()   : return  -13
    elif particle.particle_def == pp.particle.NuMuDef.get()     : return   14
    elif particle.particle_def == pp.particle.NuMuBarDef.get()  : return  -14
    elif particle.particle_def == pp.particle.TauMinusDef.get() : return   15
    elif particle.particle_def == pp.particle.TauPlusDef.get()  : return  -15
    elif particle.particle_def == pp.particle.NuTauDef.get()    : return   16
    elif particle.particle_def == pp.particle.NuTauBarDef.get() : return  -16
    elif particle.particle_def == pp.particle.Pi0Def.get()      : return  111
    elif particle.particle_def == pp.particle.PiPlusDef.get()   : return  211
    elif particle.particle_def == pp.particle.PiMinusDef.get()  : return -211
    elif particle.particle_def == pp.particle.K0Def.get()       : return  130
    elif particle.particle_def == pp.particle.KPlusDef.get()    : return  310
    elif particle.particle_def == pp.particle.KMinusDef.get()   : return -310
    elif particle.particle_def == pp.particle.PPlusDef.get()    : return 2212
    elif particle.particle_def == pp.particle.PMinusDef.get()   : return-2212
    else: return None

def is_em_primary (particle):
    """
    Given a PROPOSAL particle object as an input, returns True if the particle
    can be an electromagnetic shower primary and False otherwise
    """
    if particle.particle_def == pp.particle.EMinusDef.get(): return True
    elif particle.particle_def == pp.particle.EPlusDef.get(): return True
    elif particle.particle_def == pp.particle.GammaDef.get(): return True
    else:
       return False

def is_had_primary(particle):
    """
    Given a PROPOSAL particle object as an input, returns True if the particle
    can be a hadronic shower primary and Fasle otherwise
    """
    if particle.particle_def == pp.particle.PMinusDef.get(): return True
    elif particle.particle_def == pp.particle.PPlusDef.get(): return True
    elif particle.particle_def == pp.particle.Pi0Def.get(): return True
    elif particle.particle_def == pp.particle.PiMinusDef.get(): return True
    elif particle.particle_def == pp.particle.PiPlusDef.get(): return True
    elif particle.particle_def == pp.particle.K0Def.get(): return True
    elif particle.particle_def == pp.particle.KMinusDef.get(): return True
    elif particle.particle_def == pp.particle.KPlusDef.get(): return True
    else: return False

def is_shower_primary(particle):
    """
    Given a PROPOSAL particle object, returns True if the particle can be
    a shower primary and False otherwise
    """
    return is_em_primary(particle) or is_had_primary(particle)

"""
Codes for the DynamicData class from PROPOSAL. These represent interactions
calculated by PROPOSAL, and although most of them correspond to actual particles -
Brems is a bremsstrahlung photon, DeltaE is an ionised electron, and EPair is an
electron-positron pair, it is useful to treat them as separate entities so that
we know they come from an interaction. PROPOSAL returns particles as decay products
only, in our case.

We have followed the PDG recommendation and used numbers between 80 and 89 for
our own-defined particles.
"""
datatype_code = {
    'Data.Brems'   : 81,
    'Data.DeltaE'  : 82,
    'Data.EPair'   : 83,
    'Data.Hadrons' : 84,
    'Data.NuclInt' : 85
}

em_datatypes = [
    'Data.Brems',
    'Data.DeltaE',
    'Data.EPair'
]

hadrons_datatypes = [
    'Data.Hadrons',
    'Data.NuclInt'
]

datatype_primaries = em_datatypes + hadrons_datatypes

# NuRadioMC internal particle names organised using the PDG codes as keys
particle_name = {
        0 : 'gamma',
       11 : 'e-',
      -11 : 'e+',
       12 : 'nu_e',
      -12 : 'nu_e_bar',
       13 : 'mu-',
      -13 : 'mu+',
       14 : 'nu_mu',
      -14 : 'nu_mu_bar',
       15 : 'tau-',
      -15 : 'tau+',
       16 : 'nu_tau',
      -16 : 'nu_tau_bar',
       81 : 'brems',
       82 : 'ionized_e',
       83 : 'e_pair',
       84 : 'hadrons',
       85 : 'nucl_int',
       86 : 'decay_bundle',
      111 : 'pi0',
      211 : 'pi+',
     -211 : 'pi-',
      130 : 'K0',
      310 : 'K+',
     -310 : 'K-',
     2212 : 'p+',
    -2212 : 'p-'
}

class ProposalFunctions:
    """
    This class serves as a container for PROPOSAL functions. The functions that
    start with double underscore take PROPOSAL units as an argument and should
    not be used from the outside to avoid mismatching units.
    """

    def __create_propagator(self,
                            low=0.1*pp_PeV,
                            particle_code=13,
                            config_file='SouthPole'):
        """
        Creates a PROPOSAL propagator for muons or taus

        Parameters
        ----------
        low: float
            Minimum energy that a particle can have. If this energy is attained,
            propagation stops. In PROPOSAL units (MeV)
        particle_code: integer
            Particle code for the muon- (13), muon+ (-13), tau- (15), or tau+ (-15)
        config_file: string or path
            The user can specify the path to their own config file or choose among
            the three available options:
            -'SouthPole', a config file for the South Pole (spherical Earth). It
            consists of a 2.7 km deep layer of ice, bedrock below and air above.
            -'MooresBay', a config file for Moore's Bay (spherical Earth). It
            consists of a 576 m deep ice layer with a 2234 m deep water layer below,
            and bedrock below that.
            -'InfIce', a config file with a medium of infinite ice
            -'Greenland', a config file for Summit Station, Greenland (spherical Earth),
            same as SouthPole but with a 3 km deep ice layer.
            IMPORTANT: If these options are used, the code is more efficient if the
            user requests their own "path_to_tables" and "path_to_tables_readonly",
            pointing them to a writable directory

        Returns
        -------
        propagator: PROPOSAL propagator
            Propagator that can be used to calculate the interactions of a muon or tau
        """
        mu_def_builder = pp.particle.ParticleDefBuilder()
        if (particle_code == 13):
            mu_def_builder.SetParticleDef(pp.particle.MuMinusDef.get())
        elif (particle_code == -13):
            mu_def_builder.SetParticleDef(pp.particle.MuPlusDef.get())
        elif (particle_code == 15):
            mu_def_builder.SetParticleDef(pp.particle.TauMinusDef.get())
        elif (particle_code == -15):
            mu_def_builder.SetParticleDef(pp.particle.TauPlusDef.get())
        else:
            error_str = "The propagation of this particle via PROPOSAL is not currently supported.\n"
            error_str += "Please choose between -/+muon (13/-13) and -/+tau (15/-15)"
            raise NotImplementedError(error_str)

        mu_def_builder.SetLow(low)
        mu_def = mu_def_builder.build()

        if (config_file == 'SouthPole'):
            config_file_full_path = os.path.join(os.path.dirname(__file__), 'config_PROPOSAL.json')
        elif (config_file == 'MooresBay'):
            config_file_full_path = os.path.join(os.path.dirname(__file__), 'config_PROPOSAL_mooresbay.json')
        elif (config_file == 'InfIce'):
            config_file_full_path = os.path.join(os.path.dirname(__file__), 'config_PROPOSAL_infice.json')
        elif (config_file == 'Greenland'):
            config_file_full_path = os.path.join(os.path.dirname(__file__), 'config_PROPOSAL_greenland.json')
        elif (os.path.exists(config_file)):
            config_file_full_path = config_file
        else:
            raise ValueError("Proposal config file is not valid. Please provide a valid option.")

        if not os.path.exists(config_file_full_path):
            error_message  = "Proposal config file does not exist.\n"
            error_message += "Please provide valid paths for the interpolation tables "
            error_message += "in file {}.sample ".format(config_file_full_path)
            error_message += "and copy the file to {}.".format(os.path.basename(config_file_full_path))
            raise ValueError(error_message)

        propagator = pp.Propagator(particle_def=mu_def, config_file=config_file_full_path)

        return propagator

    def __get_compact_sub_pev_losses(self,
                                     energy_arr,
                                     distance_arr,
                                     compact_dist,
                                     min_energy_loss):
        r""" return biggest compact loss if above min_energy_cut

        This function groups energy losses along a path and groups them into a
        single shower. The effect is only seen for < 10 PeV energy bins and it's
        at least one order of magnitude lower than non-grouped losses, so it does
        not influence that much.

        Parameters
        ----------
        energy_arr: array-like
            energy of the energy losses below min_energy_loss, in PROPOSAL units (MeV)
        distance_arr: array_like
            distances of the energy losses below min_energy_loss, in PROPOSAL units (cm)
        compact_dist: float
            distance in centimeters (PROPOSAL units): how compact the energy losses should be
        min_energy_loss: float
            min energy for the sensitivity (here a PeV), in PROPOSAL units (MeV)
        """
        len_bins = np.arange(distance_arr[0], distance_arr[-1] + 1e-3, 100)
        # We have used 100 to create a bin length of 1 m
        len_indices = np.digitize(distance_arr, len_bins)
        bincount = np.bincount(len_indices, energy_arr)

        if len(bincount) <= compact_dist:
            sum_bins = np.sum(bincount)
            if sum_bins > min_energy_loss:
                return [sum_bins]
            else:
                return []

        # We transform the compact_dist into meters, since the above histogram
        # has a bin length of 1 m
        convolved_comp_arr = np.convolve(np.ones(int(compact_dist/pp_m)), bincount, mode='valid')
        if np.any(convolved_comp_arr > min_energy_loss):
            return [np.max(convolved_10m_arr)]
        else:
            return []

    def __produces_shower(self,
                          particle,
                          min_energy_loss=1*pp_PeV):
        """
        Returns True if the input particle or interaction can be a shower primary
        and its energy is above min_energy_loss

        Parameters
        ----------
        particle: PROPOSAL particle or DynamicData (interaction)
            Input particle
        min_energy_loss: float
            Threshold above which a particle shower is considered detectable
            or relevant, in PROPOSAL units (MeV)

        Returns
        -------
        bool
            True if particle produces shower, False otherwise
        """
        energy_threshold = particle.energy > min_energy_loss
        if not energy_threshold:
            return False

        if particle.id == pp.particle.Data.Particle:
            shower_inducing = is_shower_primary(particle)
        elif str(particle.id) in datatype_primaries:
            shower_inducing = True
        else:
            return False

        return shower_inducing

    def __shower_properties(self, particle):
        """
        Calculates shower properties for the shower created by the input particle

        Parameters
        ----------
        particle: PROPOSAL particle or DynamicData

        Returns
        -------
        shower_type: string
            'em' for EM showers and 'had' for hadronic showers
        code: integer
            Particle code for the shower primary
        name: string
            Name of the shower primary
        """
        if particle.id == pp.particle.Data.Particle:
            if is_em_primary(particle):
                shower_type = 'em'
            elif is_had_primary(particle):
                shower_type = 'had'
            else:
                return None, None, None

            code = particle_code(particle)

        elif str(particle.id) in em_datatypes:
            shower_type = 'em'
            code = datatype_code[str(particle.id)]

        elif str(particle.id) in hadrons_datatypes:
            shower_type = 'had'
            code = datatype_code[str(particle.id)]

        name = particle_name[code]

        return shower_type, code, name

    def __propagate_particle(self,
                             energy_lepton,
                             lepton_code,
                             lepton_position,
                             lepton_direction,
                             propagation_length,
                             propagators):
        """
        Calculates secondary particles using a PROPOSAL propagator. It needs to
        be given a propagators dictionary with particle codes as key

        Parameters
        ----------
        energy_lepton: float
            Energy of the input lepton, in PROPOSAL units (MeV)
        lepton_code: integer
            Input lepton code
        lepton_position: (float, float, float) tuple
            Position of the input lepton, in PROPOSAL units (cm)
        lepton_direction: (float, float, float) tuple
            Lepton direction vector, normalised to 1
        propagation_length: float
            Maximum length the particle is propagated, in PROPOSAL units (cm)

        Returns
        -------
        secondaries: array of PROPOSAL particles
            Secondaries created by the propagation
        """
        x, y, z = lepton_position
        px, py, pz = lepton_direction
        propagators[lepton_code].particle.position = pp.Vector3D(x, y, z)
        propagators[lepton_code].particle.direction = pp.Vector3D(px, py, pz)
        propagators[lepton_code].particle.propagated_distance = 0
        propagators[lepton_code].particle.energy = energy_lepton

        secondaries = propagators[lepton_code].propagate(propagation_length)

        return secondaries

    def __filter_secondaries(self,
                             secondaries,
                             min_energy_loss,
                             lepton_position):
        """
        Takes an input secondary particles array and returns an array with the
        SecondaryProperties of those particles that create a shower above a threshold

        Parameters
        ----------
        secondaries: array of PROPOSAL particles
            Array with secondary particles
        min_energy_loss: float
            Threshold for shower production, in PROPOSAL units (MeV)
        lepton_position: tuple (float, float, float)
            Initial position of the primary lepton, in PROPOSAL units (cm)

        Returns
        -------
        shower_inducing_prods: list
            List containing the secondary properties of every shower-inducing
            secondary particle, in NuRadioMC units
        """

        shower_inducing_prods = []

        for sec in secondaries:

            # Decays contain only one shower-inducing particle
            # Muons and neutrinos resulting from decays are ignored
            if self.__produces_shower(sec, min_energy_loss):

                distance  = ( (sec.position.x - lepton_position[0]) * units.cm )**2
                distance += ( (sec.position.y - lepton_position[1]) * units.cm )**2
                distance += ( (sec.position.z - lepton_position[2]) * units.cm )**2
                distance  = np.sqrt(distance)
                energy = sec.energy * units.MeV

                shower_type, code, name = self.__shower_properties(sec)

                shower_inducing_prods.append( SecondaryProperties(distance, energy, shower_type, code, name) )

        return shower_inducing_prods

    def get_secondaries_array(self,
                              energy_leptons_nu,
                              lepton_codes,
                              lepton_positions_nu=None,
                              lepton_directions=None,
                              config_file='SouthPole',
                              low_nu=0.1*units.PeV,
                              propagation_length_nu=1000*units.km,
                              min_energy_loss_nu=1*units.PeV,
                              propagate_decay_muons=True):
        """
        Propagates a set of leptons and returns a list with the properties for
        all the properties of the shower-inducing secondary particles

        Parameters
        ----------
        energy_leptons_nu: array of floats
            Array with the energies of the input leptons, in NuRadioMC units (eV)
        lepton_codes: array of integers
            Array with the PDG lepton codes
        lepton_positions_nu: array of (float, float, float) tuples
            Array containing the lepton positions in NuRadioMC units (m)
        lepton_directions: array of (float, float, float) tuples
            Array containing the lepton directions, normalised to 1
        config_file: string or path
            The user can specify the path to their own config file or choose among
            the three available options:
            -'SouthPole', a config file for the South Pole (spherical Earth)
            -'MooresBay', a config file for Moore's Bay (spherical Earth)
            -'InfIce', a config file with a medium of infinite ice
            -'Greenland', a config file for Summit Station, Greenland (spherical Earth)
            IMPORTANT: If these options are used, the code is more efficient if the
            user requests their own "path_to_tables" and "path_to_tables_readonly",
            pointing them to a writable directory
        low_nu: float
            Low energy limit for the propagating particle in NuRadioMC units (eV)
        propagation_length_nu: float
            Maximum propagation length in NuRadioMC units (m)
        min_energy_loss_nu: float
            Minimum energy for a selected secondary-induced shower (eV)
        propagate_decay_muons: bool
            If True, muons created by tau decay are propagated and their induced
            showers are stored

        Returns
        -------
        secondaries_array: 2D-list containing SecondaryProperties objects
            List containing the information on the shower-inducing secondaries. The
            first dimension indicates the primary lepton and the second dimension
            navigates through the secondaries produced by that primary.

            The SecondaryProperties objects are expressed in NuRadioMC units.
        """

        # Converting to PROPOSAL units
        low = low_nu * pp_eV
        propagation_length = propagation_length_nu * pp_m
        min_energy_loss = min_energy_loss_nu * pp_eV
        energy_leptons = np.array(energy_leptons_nu) * pp_eV

        if lepton_positions_nu is None:
            lepton_positions = [(0, 0, 0)] * len(energy_leptons)
        else:
            lepton_positions = [ np.array(lepton_position_nu) * pp_m for lepton_position_nu in lepton_positions_nu ]

        if lepton_directions is None:
            lepton_directions = [(0, 0, -1)] * len(energy_leptons)

        propagators = {}
        for lepton_code in np.unique(lepton_codes):
            if lepton_code not in propagators:
                propagators[lepton_code] = self.__create_propagator(low=low, particle_code=lepton_code,
                                                                    config_file=config_file)

        if propagate_decay_muons:
            # We create another muon propagator dictionary to try to avoid a segmentation
            # fault happening in some installations of Proposal
            mu_propagators = {}
            for muon_code in [13, -13]:
                mu_propagators[muon_code] = self.__create_propagator(low=low, particle_code=muon_code,
                                                                     config_file=config_file)
            decay_muons_array = []

        secondaries_array = []

        for energy_lepton, lepton_code, lepton_position, lepton_direction in zip(energy_leptons,
            lepton_codes, lepton_positions, lepton_directions):

            secondaries = self.__propagate_particle(energy_lepton, lepton_code,
                                                    lepton_position, lepton_direction,
                                                    propagation_length, propagators)

            shower_inducing_prods = self.__filter_secondaries(secondaries, min_energy_loss, lepton_position)

            # Checking if there is a muon in the products
            if propagate_decay_muons:

                decay_muons_array.append([None, None, None, None])

                for sec in secondaries:

                    if (sec.id != pp.particle.Data.Particle):
                        continue

                    if (sec.particle_def == pp.particle.MuMinusDef.get()) or (sec.particle_def == pp.particle.MuPlusDef.get()):

                        if sec.particle_def == pp.particle.MuMinusDef.get():
                            mu_code = 13
                        elif sec.particle_def == pp.particle.MuPlusDef.get():
                            mu_code = -13

                        mu_energy = sec.energy
                        if (mu_energy <= low):
                            continue
                        mu_position = (sec.position.x, sec.position.y, sec.position.z)
                        mu_direction = lepton_direction # We reuse the primary lepton direction
                                                        # At these energies the muon direction is the same
                        decay_muons_array[-1] = [mu_energy, mu_code, mu_position, mu_direction]
                        # I store the properties of each muon in an array because they cannot be
                        # propagated while we are looping the secondaries array. Doing that can
                        # create a segmentation fault because the results of the new propagation
                        # are written into the secondaries array (!)

            # group shower-inducing decay products so that they create a single shower
            min_distance = 0.1 * units.m
            while( len(shower_inducing_prods) > 1 and
                   np.abs(shower_inducing_prods[-1].distance - shower_inducing_prods[-2].distance) < min_distance):

                last_decay_prod = shower_inducing_prods.pop(-1)
                shower_inducing_prods[-1].energy += last_decay_prod.energy
                shower_inducing_prods[-1].code = 86
                shower_inducing_prods[-1].name = particle_name[86]

            secondaries_array.append(shower_inducing_prods)

        # Propagating the decay muons
        if propagate_decay_muons:

            for shower_inducing_prods, decay_muon, lepton_position in zip(secondaries_array,
                decay_muons_array, lepton_positions):

                if decay_muon[0] is None:
                    continue
                mu_energy, mu_code, mu_position, mu_direction = decay_muon
                mu_secondaries = self.__propagate_particle(mu_energy, mu_code, mu_position, mu_direction,
                                                           propagation_length, mu_propagators)

                mu_shower_inducing_prods = self.__filter_secondaries(mu_secondaries, min_energy_loss, lepton_position)

                shower_inducing_prods += mu_shower_inducing_prods

        return secondaries_array

    def get_decays(self,
                   energy_leptons_nu,
                   lepton_codes,
                   lepton_positions_nu = None,
                   lepton_directions = None,
                   config_file='InfIce',
                   low_nu=0.1*units.PeV,
                   propagation_length_nu=1000*units.km):
        """
        Propagates a set of leptons and returns a list with the properties of
        the decay particles.

        Parameters
        ----------
        energy_leptons_nu: array of floats
            Array with the energies of the input leptons, in NuRadioMC units (eV)
        lepton_codes: array of integers
            Array with the PDG lepton codes
        lepton_positions_nu: array of (float, float, float) tuples
            Array containing the lepton positions in NuRadioMC units (m)
        lepton_directions: array of (float, float, float) tuples
            Array containing the lepton directions, normalised to 1
        config_file: string or path
            The user can specify the path to their own config file or choose among
            the three available options:
            -'SouthPole', a config file for the South Pole (spherical Earth)
            -'MooresBay', a config file for Moore's Bay (spherical Earth)
            -'InfIce', a config file with a medium of infinite ice
            -'Greenland', a config file for Summit Station, Greenland (spherical Earth)
            IMPORTANT: If these options are used, the code is more efficient if the
            user requests their own "path_to_tables" and "path_to_tables_readonly",
            pointing them to a writable directory
        low_nu: float
            Low energy limit for the propagating particle in NuRadioMC units (eV)
        propagation_length_nu: float
            Maximum propagation length in NuRadioMC units (m)

        Returns
        -------
        decays_array: array of (float, float) tuples
            The first element of the tuple contains the decay distance in m
            The second element contains the decay energy in eV (NuRadioMC units)
        """

        # Converting to PROPOSAL units
        low = low_nu * pp_eV
        propagation_length = propagation_length_nu * pp_m
        energy_leptons = np.array(energy_leptons_nu) * pp_eV

        if lepton_positions_nu is None:
            lepton_positions = [(0, 0, 0)] * len(energy_leptons)
        else:
            lepton_positions = [ np.array(lepton_position_nu) * pp_m for lepton_position_nu in lepton_positions_nu ]

        if lepton_directions is None:
            lepton_directions = [(0, 0, 1)] * len(energy_leptons)

        propagators = {}
        for lepton_code in lepton_codes:
            if lepton_code not in propagators:
                propagators[lepton_code] = self.__create_propagator(low=low, particle_code=lepton_code,
                                                                    config_file=config_file)

        decays_array = []

        for energy_lepton, lepton_code, lepton_position, lepton_direction in zip(energy_leptons,
            lepton_codes, lepton_positions, lepton_directions):

            decay_prop = (None, None)

            while( decay_prop == (None,None) ):

                secondaries = self.__propagate_particle(energy_lepton, lepton_code, lepton_position, lepton_direction,
                                                        propagation_length, propagators)

                decay_particles = np.array([p for p in secondaries if p.id == pp.particle.Data.Particle])
                decay_energies = np.array([p.energy for p in decay_particles])
                decay_energy = np.sum(decay_energies) * units.MeV

                try:
                    # If Proposal fails and there is no decay (sometimes it happens),
                    # the particle is propagated again
                    decay_distance  = ( (decay_particles[0].position.x - lepton_position[0]) * units.cm )**2
                    decay_distance += ( (decay_particles[0].position.y - lepton_position[1]) * units.cm )**2
                    decay_distance += ( (decay_particles[0].position.z - lepton_position[2]) * units.cm )**2
                    decay_distance  = np.sqrt(decay_distance)

                    decay_prop = (decay_distance, decay_energy)
                    decays_array.append(decay_prop)
                except:
                    decay_prop = (None, None)

        return np.array(decays_array)



import glob
import os
import sys
import numpy as np
from collections import OrderedDict
import h5py
import argparse
import logging
logger = logging.getLogger("HDF5-merger")
logging.basicConfig(level=logging.DEBUG)
logger.setLevel(logging.WARNING)


def merge2(filenames, output_filename):
    data = OrderedDict()
    attrs = OrderedDict()
    groups = OrderedDict()
    group_attrs = OrderedDict()
    n_data = {}
    n_groups = {}
    non_empty_filenames = []
    n_events_total = 0

    for f in filenames:
        logger.info("adding file {}".format(f))
        fin = h5py.File(f, 'r')
        n_events_total += fin.attrs['n_events']
        logger.debug(f"increasing total number of events by {fin.attrs['n_events']:d} to {n_events_total:d} ")

        if(np.sum(np.array(fin['triggered'])) == 0):
            logger.info(f"file {f} contains no events")
        else:
            non_empty_filenames.append(f)
            logger.debug(f"file {f} contains {np.sum(np.array(fin['triggered']))} triggered events.")

        data[f] = {}
        groups[f] = {}

        for key in fin:
            if isinstance(fin[key], h5py._hl.group.Group):
                groups[f][key] = {}
                if(key not in n_groups):
                    n_groups[key] = {}
                for key2 in fin[key]:
                    groups[f][key][key2] = fin[key][key2][...]
                    if(key2 not in n_groups[key]):
                        n_groups[key][key2] = 0
                    n_groups[key][key2] += len(groups[f][key][key2])
                if(key not in group_attrs):
                    group_attrs[key] = {}
                    for key2 in fin[key].attrs:
                        group_attrs[key][key2] = fin[key].attrs[key2]
                else:
                    for key2 in fin[key].attrs:
                        if(not np.all(group_attrs[key][key2] == fin[key].attrs[key2])):
                            logger.warning(f"attribute {key2} of group {key} of file {filenames[0]} and {f} are different ({group_attrs[key][key2]} vs. {fin[key].attrs[key2]}. Using attribute value of first file, but you have been warned!")
            else:
                data[f][key] = fin[key][...]
                if(key not in n_data):
                    n_data[key] = 0
                n_data[key] += len(data[f][key])

        for key in fin.attrs:
            if(key not in attrs):
                attrs[key] = fin.attrs[key]
            else:
                if(key != 'trigger_names'):
                    if(not np.all(attrs[key] == fin.attrs[key])):
                        if(key == "n_events"):
                            logger.warning(f"number of events in file {filenames[0]} and {f} are different ({attrs[key]} vs. {fin.attrs[key]}. We keep track of the total number of events, but in case the simulation was performed with different settings per file (e.g. different zenith angle bins), the averaging might be effected.")
                        else:
                            logger.warning(f"attribute {key} of file {filenames[0]} and {f} are different ({attrs[key]} vs. {fin.attrs[key]}. Using attribute value of first file, but you have been warned!")
            if((('trigger_names' not in attrs) or (len(attrs['trigger_names']) == 0)) and 'trigger_names' in fin.attrs):
                attrs['trigger_names'] = fin.attrs['trigger_names']
        fin.close()

    # create data sets
    logger.info("creating data sets")
    fout = h5py.File(output_filename, 'w')
    if(len(non_empty_filenames)):
        keys = data[non_empty_filenames[0]]
        for key in keys:
            logger.info(f"merging key {key}")
            all_files_have_key = True
            for f in non_empty_filenames:
                if(not key in data[f]):
                    logger.debug(f"key {key} not in {f}")
                    all_files_have_key = False
            if(not all_files_have_key):
                logger.warning(f"not all files have the key {key}. This key will not be present in the merged file.")
                continue
            shape = list(data[non_empty_filenames[0]][key].shape)
            shape[0] = n_data[key]

            tmp = np.zeros(shape, dtype=data[non_empty_filenames[0]][key].dtype)

            i = 0
            for f in non_empty_filenames:
                tmp[i:(i + len(data[f][key]))] = data[f][key]
                i += len(data[f][key])

            fout.create_dataset(key, tmp.shape, dtype=tmp.dtype,
                                compression='gzip')[...] = tmp

        keys = groups[non_empty_filenames[0]]
        for key in keys:
            logger.info("writing group {}".format(key))
            g = fout.create_group(key)
            for key2 in groups[non_empty_filenames[0]][key]:
                logger.info("writing data set {}".format(key2))
                all_files_have_key = True
                for f in non_empty_filenames:
                    if(not key2 in groups[f][key]):
                        logger.debug(f"key {key2} of group {key} not in {f}")
                        all_files_have_key = False
                if(not all_files_have_key):
                    logger.warning(f"not all files have the key {key2}. This key will not be present in the merged file.")
                    continue

                shape = list(groups[non_empty_filenames[0]][key][key2].shape)
                shape[0] = n_groups[key][key2]

                tmp = np.zeros(shape, dtype=groups[non_empty_filenames[0]][key][key2].dtype)
                i = 0
                for f in non_empty_filenames:
                    tmp[i:(i + len(groups[f][key][key2]))] = groups[f][key][key2]
                    i += len(groups[f][key][key2])

                g.create_dataset(key2, shape, dtype=groups[non_empty_filenames[0]][key][key2].dtype,
                                 compression='gzip')[...] = tmp
            # save group attributes
            for key2 in group_attrs[key]:
                fout[key].attrs[key2] = group_attrs[key][key2]
        # save all atrributes
        attrs['n_events'] = n_events_total
        for key in attrs:
            fout.attrs[key] = attrs[key]
    else:  # now handle the case
        logger.warning("All files are empty. Copying content of first file to output file and keepting track of total number of simulated events.")
        # all files are empty, so just copy the content of the first file (attributes and empyt data sets) to the output file
        # update n_events attribute with the total number of events
        fin = h5py.File(filenames[0], 'r')
        for key in fin.attrs:
            if(key == "n_events"):
                fout.attrs[key] = n_events_total
            else:
                fout.attrs[key] = fin.attrs[key]
        for key in fin:
            if isinstance(fin[key], h5py._hl.group.Group):
                g = fout.create_group(key)
                for key2 in fin[key]:
                    g.create_dataset(key2, fin[key][key2].shape, dtype=fin[key][key2].dtype,
                                     compression='gzip')[...] = fin[key][key2]
                for key2 in fin[key].attrs:
                    g.attrs[key2] = fin[key].attrs[key2]
            else:
                fout.create_dataset(key, fin[key].shape, dtype=fin[key].dtype,
                                    compression='gzip')[...] = fin[key]

#     # save all data to hdf5
#     for key in data[filenames[0]]:
#         print("writing data set {}".format(key))
#         i = 0
#         for f in data:
#             fout[key][i:(i+len(data[f][key]))] = data[f][key]
#             i += len(data[f][key])
    # save all group data to hdf5
#     for key in groups[filenames[0]]:
#         print("writing group {}".format(key))
#         for key2 in groups[filenames[0]][key]:
#             print("writing data set {}".format(key2))
#             i = 0
#             for f in groups:
#                 fout[key][key2][i:(i+len(groups[f][key][key2]))] = groups[f][key][key2]
#                 i += len(groups[f][key][key2])
#         # save group attributes
#         for key2 in group_attrs[key]:
#             fout[key].attrs[key2] = group_attrs[key][key2]
#

    fout.close()


if __name__ == "__main__":
    """
    merges multiple hdf5 output files into one single files.
    The merger module automatically keeps track of the total number
    of simulated events (which are needed to correctly calculate the effective volume).

    The script expects that the folder structure is
    ../output/energy/*.hdf5.part????

    Optional log level setting to either set DEBUG, INFO, or WARNING to the readout. Example: add --loglevel DEBUG when calling script to set loglevel to DEBUG. 
    """
    parser = argparse.ArgumentParser(description='Merge hdf5 files')
    parser.add_argument('files', nargs='+', help='input file or files')
    parser.add_argument('--loglevel', metavar='level', help='loglevel set to either DEBUG, INFO, or WARNING')
    args = parser.parse_args()

    if args.loglevel is not None:
        log_val = eval(f'logging.{args.loglevel}')
        logger.setLevel(log_val)

    if(len(args.files) < 1):
        print("usage: python merge_hdf5.py /path/to/simulation/output/folder\nor python merge_hdf5.py outputfilename input1 input2 ...")
    elif(len(args.files) == 1):
        filenames = glob.glob("{}/*/*.hdf5.part????".format(args.files[0]))
        filenames2 = []
        for i, filename in enumerate(filenames):
            filename, ext = os.path.splitext(filename)
            if(ext != '.hdf5'):
                if(filename not in filenames2):
                    d = os.path.split(filename)
                    a, b = os.path.split(d[0])
                    filenames2.append(filename)

        for filename in filenames2:
            if(os.path.splitext(filename)[1] == '.hdf5'):
                d = os.path.split(filename)
                a, b = os.path.split(d[0])
                output_filename = os.path.join(a, d[1])  # remove subfolder from filename
                if(os.path.exists(output_filename)):
                    logger.error('file {} already exists, skipping'.format(output_filename))
                else:
                    #                 try:
                    input_files = np.array(sorted(glob.glob(filename + '.part????')))
                    mask = np.array([os.path.getsize(x) > 1000 for x in input_files], dtype=np.bool)
                    if(np.sum(~mask)):
                        logger.warning("{:d} files were deselected because their filesize was to small".format(np.sum(~mask)))

                    merge2(input_files[mask], output_filename)
    #                 except:
    #                     print("failed to merge {}".format(filename))
    elif(len(args.files) > 1):
        output_filename = args.files[0]
        if(os.path.exists(output_filename)):
            logger.error('file {} already exists, skipping'.format(output_filename))
        else:
            input_files = args.files[1:]
            merge2(input_files, output_filename)



import numpy as np
from NuRadioReco.utilities import units
from matplotlib import pyplot as plt
from radiotools import helper as hp
from NuRadioMC.utilities import cross_sections as cs
from NuRadioMC.utilities import earth_attenuation
from scipy.integrate import quad
from scipy.optimize import brentq
from scipy.interpolate import interp1d
from NuRadioMC.simulation.simulation import pretty_time_delta
from NuRadioMC.EvtGen.generator import write_events_to_hdf5
from NuRadioMC.utilities import inelasticities
import pickle
import os
import time
import logging
import NuRadioMC
from NuRadioMC.utilities import version
# np.random.seed(10)  # just for testing

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

R_earth = 6357390 * units.m
earth = earth_attenuation.PREM()


def generate_eventlist_cylinder(filename, n_events, Emin, Emax,
                                full_rmin=None, full_rmax=None, full_zmin=None, full_zmax=None,
                                thetamin=0.*units.rad, thetamax=np.pi * units.rad,
                                phimin=0.*units.rad, phimax=2 * np.pi * units.rad,
                                start_event_id=1,
                                flavor=[12, -12, 14, -14, 16, -16],
                                n_events_per_file=None,
                                spectrum='log_uniform',
                                start_file_id=0):
    """
    Event generator

    Generates neutrino interactions, i.e., vertex positions, neutrino directions,
    neutrino flavor, charged currend/neutral current and inelastiviy distributions.
    All events are saved in an hdf5 file.

    Parameters
    ----------
    filename: string
        the output filename of the hdf5 file
    n_events: int
        number of events to generate
    Emin: float
        the minimum neutrino energy (energies are randomly chosen assuming a
        uniform distribution in the logarithm of the energy)
    Emax: float
        the maximum neutrino energy (energies are randomly chosen assuming a
        uniform distribution in the logarithm of the energy)

    full_rmin: float (default None)
        lower r coordinate of simulated volume (if None it is set to 1/3 of the fiducial volume, if second vertices are not activated it is set to the fiducial volume)
    full_rmax: float (default None)
        upper r coordinate of simulated volume (if None it is set to 5x the fiducial volume, if second vertices are not activated it is set to the fiducial volume)
    full_zmin: float (default None)
        lower z coordinate of simulated volume (if None it is set to 1/3 of the fiducial volume, if second vertices are not activated it is set to the fiducial volume)
    full_zmax: float (default None)
        upper z coordinate of simulated volume (if None it is set to 5x the fiducial volume, if second vertices are not activated it is set to the fiducial volume)
    thetamin: float
        lower zenith angle for neutrino arrival direction
    thetamax: float
        upper zenith angle for neutrino arrival direction
    phimin: float
        lower azimuth angle for neutrino arrival direction
    phimax: float
         upper azimuth angle for neutrino arrival direction
    start_event: int
        default: 1
        event number of first event
    flavor: array of ints
        default: [12, -12, 14, -14, 16, -16]
        specify which neutrino flavors to generate. A uniform distribution of
        all specified flavors is assumed.
        The neutrino flavor (integer) encoded as using PDF numbering scheme,
        particles have positive sign, anti-particles have negative sign,
        relevant for us are:
        * 12: electron neutrino
        * 14: muon neutrino
        * 16: tau neutrino
    n_events_per_file: int or None
        the maximum number of events per output files. Default is None, which
        means that all events are saved in one file. If 'n_events_per_file' is
        smaller than 'n_events' the event list is split up into multiple files.
        This is useful to split up the computing on multiple cores.
    spectrum: string
        defines the probability distribution for which the neutrino energies are generated
        * 'log_uniform': uniformly distributed in the logarithm of energy
    start_file_id: int (default 0)
        in case the data set is distributed over several files, this number specifies the id of the first file
        (useful if an existing data set is extended)
        if True, generate deposited energies instead of primary neutrino energies
    """

    attributes = {}
    n_events = int(n_events)

    # save current NuRadioMC version as attribute
    # save NuRadioMC and NuRadioReco versions
    attributes['NuRadioMC_EvtGen_version'] = NuRadioMC.__version__
    attributes['NuRadioMC_EvtGen_version_hash'] = version.get_NuRadioMC_commit_hash()

    attributes['start_event_id'] = start_event_id

    attributes['fiducial_rmin'] = full_rmin
    attributes['fiducial_rmax'] = full_rmax
    attributes['fiducial_zmin'] = full_zmin
    attributes['fiducial_zmax'] = full_zmax
    attributes['rmin'] = full_rmin
    attributes['rmax'] = full_rmax
    attributes['zmin'] = full_zmin
    attributes['zmax'] = full_zmax
    attributes['flavors'] = flavor
    attributes['Emin'] = Emin
    attributes['Emax'] = Emax
    attributes['thetamin'] = thetamin
    attributes['thetamax'] = thetamax
    attributes['phimin'] = phimin
    attributes['phimax'] = phimax

    # define cylinder by two points and the radius
    h_cylinder = full_zmax - full_zmin
    r_cylinder = full_rmax
    pt1 = np.array([0, 0, R_earth + full_zmax])
    pt2 = np.array([0, 0, R_earth + full_zmin])

    data_sets = {}

    # calculate maximum width of projected area
    theta_max = np.arctan(h_cylinder / 2 / r_cylinder)
    d = 2 * r_cylinder * np.cos(theta_max) + h_cylinder * np.sin(theta_max)  # width of area

    print(f"cylinder r = {r_cylinder/units.km:.1f}km, h = {h_cylinder/units.km:.1f}km -> dmax = {d/units.km:.1f}km")

    def perp(a) :
        b = np.empty_like(a)
        b[0] = -a[1]
        b[1] = a[0]
        return b

    # line segment a given by endpoints a1, a2
    # line segment b given by endpoints b1, b2
    # return
    def seg_intersect(a1, a2, b1, b2) :
        da = a2 - a1
        db = b2 - b1
        dp = a1 - b1
        dap = perp(da)
        denom = np.dot(dap, db)
        num = np.dot(dap, dp)
        return (num / denom) * db + b1

    def get_R(t, v, X):
        """"
        calculate distance to center of Earth as a function of travel distance
        
        Parameters
        -----------
        t: 3dim array
            travel distance
        v: 3dim array
            direction
        X: 3dim array
            start point
        """
        return np.linalg.norm(v * t + X)

    def get_density(t, v, X):
        """
        calculates density as a function of travel distance
        
        Parameters
        -----------
        t: 3dim array
            travel distance
        v: 3dim array
            direction
        X: 3dim array
            start point
        """
        return  earth.density(get_R(t, v, X))

    def slant_depth(t, v, X):
        """
        calculates slant depth (grammage) as a function of travel distance
        
        Parameters
        -----------
        t: 3dim array
            travel distance
        v: 3dim array
            direction
        X: 3dim array
            start point
        """
        res = quad(get_density, 0, t, args=(v, X), limit=200)
        return res[0]

    def slant_depth_num(t, v, X, step=50 * units.m):
        """
        calculates slant depth (grammage) as a function of travel distance
        
        Parameters
        -----------
        t: 3dim array
            travel distance
        v: 3dim array
            direction
        X: 3dim array
            start point
        """
        tt = np.linspace(0, t, t / step)
        rr = np.linalg.norm(X + np.outer(tt, v), axis=1)
        res = np.trapz(earth.density(rr), tt)
        return res

    def obj_dist_to_surface(t, v, X):
        return get_R(t, v, X) - R_earth

    def obj(t, v, X, Lint):
        """
        objective function to determine at which travel distance we reached the interaction point
        """
        return slant_depth(t, v, X) - Lint

    def points_in_cylinder(pt1, pt2, r, q):
        """
        determines if point lies within a cylinder
        
        Parameters
        -----------
        pt1: 3dim array
            lowest point on cylinder axis
        pt2: 3dim array
            highest point on cylinder axis
        r: float
            radius of cylinder
        q: 3dim array
            point under test
        
        Returns True/False
        """
        vec = pt2 - pt1
        const = r * np.linalg.norm(vec)
        return len(np.where(np.dot(q - pt1, vec) >= 0 and np.dot(q - pt2, vec) <= 0 and np.linalg.norm(np.cross(q - pt1, vec)) <= const)[0]) > 0

    # precalculate the maximum slant depth to the detector
    if(not os.path.exists("buffer_Llimit.pkl")):
        zens = np.arange(0, 180.1 * units.deg, 2 * units.deg)
        Lint_max = np.zeros_like(zens)
        Lint_min = np.zeros_like(zens)
        Xs = np.array([[0, -0.5 * r_cylinder, -h_cylinder + R_earth],
                       [0, -0.5 * r_cylinder, R_earth],
                       [0, 0.5 * r_cylinder, -h_cylinder + R_earth],
                       [0, 0.5 * r_cylinder, R_earth]]
                       )
        v_tmps = -hp.spherical_to_cartesian(zens, np.zeros_like(zens))  # neutrino direction
        for i in range(len(v_tmps)):
            v = v_tmps[i]
            sdepth_tmp = np.zeros(4)
            for j, X in enumerate(Xs):
                if((X[2] == R_earth) and (zens[i] <= 90 * units.deg)):
                    t = 0
                    sdepth_tmp[j] = 0
                else:
                    t = brentq(obj_dist_to_surface, 100, 2 * R_earth, args=(-v, X))
                    sdepth_tmp[j] = slant_depth_num(t, -v, X)
        #         print(i, zens[i] / units.deg, X, sdepth_tmp[j])
        #     enter_point = X + (-v * t)
            Lint_max[i] = np.max(sdepth_tmp)
            Lint_min[i] = np.min(sdepth_tmp)
        pickle.dump([zens, Lint_max, Lint_min], open("buffer_Llimit.pkl", "wb"), protocol=4)
    else:
        zens, Lint_max, Lint_min = pickle.load(open("buffer_Llimit.pkl", "rb"))

    get_Lmax = interp1d(zens, Lint_max, kind='next')
    get_Lmin = interp1d(zens, Lint_min, kind='previous')

    if 0:
        fig, a = plt.subplots(1, 1)
        ztmp = np.linspace(0, 180 * units.deg, 10000)
        a.plot(ztmp / units.deg, get_Lmax(ztmp) / units.g * units.cm ** 2, 'C0-', label="max possible Lint")
    #     a.plot(zens / units.deg, Lint_max / units.g * units.cm ** 2, 'oC0')
        a.plot(ztmp / units.deg, get_Lmin(ztmp) / units.g * units.cm ** 2, 'C1-', label="min possible Lint")
    #     a.plot(zens / units.deg, Lint_min / units.g * units.cm ** 2, 'dC1')
        a.hlines(cs.get_interaction_length(.1 * units.EeV, 1, 12, "total") / units.g * units.cm ** 2, 0, 180, label="0.1 EeV", colors='C2')
        a.hlines(cs.get_interaction_length(1 * units.EeV, 1, 12, "total") / units.g * units.cm ** 2, 0, 180 , label="1 EeV", colors='C3')
        a.hlines(cs.get_interaction_length(10 * units.EeV, 1, 12, "total") / units.g * units.cm ** 2, 0, 180 , label="10 EeV", colors='C4')
        a.set_xlabel("zenith angle [deg]")
        a.set_ylabel("slant depth [g/cm^2]")
        a.semilogy(True)
        a.set_xticks(np.arange(0, 181, 10))
        a.set_ylim(5e5)
        a.legend()

        fig.tight_layout()
        fig.savefig("Lvszen.png")
        plt.show()

    failed = 0
    if(spectrum == 'log_uniform'):
        Enu = 10 ** np.random.uniform(np.log10(Emin), np.log10(Emax), n_events)
    flavors = np.array([flavor[i] for i in np.random.randint(0, high=len(flavor), size=n_events)])
    az = np.random.uniform(phimin, phimax, n_events)
    zen = np.arccos(np.random.uniform(np.cos(thetamax), np.cos(thetamin), n_events))
    # generate random positions on an area perpendicular do neutrino direction
    ax, ay = np.random.uniform(-0.5 * d, 0.5 * d, (2, n_events))
    # az = np.ones(n_events) * (R_earth - .5 * h_cylinder)  # move plane to the center of the cylinder

    # calculate grammage (g/cm^2) after which neutrino interacted
    Lint = np.random.exponential(cs.get_interaction_length(Enu, 1, flavors, "total"), n_events)

    mask = (Lint < get_Lmax(zen)) & (Lint > get_Lmin(zen))
    print(f"{np.sum(mask)}/{n_events} = {np.sum(mask)/n_events:.2g} can potentially interact in simulation volume")

    # calculate position where neutrino interacts
    data_sets = {'xx': [],
                'yy': [],
                'zz': [],
                'azimuths': [],
                'zeniths': [],
                'flavors': [],
                'energies': []}
    # calculate rotation matrix to transform position on area to 3D
    mask_int = np.zeros_like(mask, dtype=np.bool)
    t0 = time.perf_counter()
    n_cylinder = 0
    for j, i in enumerate(np.arange(n_events, dtype=np.int)[mask]):
        if(j % 1000 == 0 and i > 0):
            eta = (time.perf_counter() - t0) * (n_events - i) / i
            logger.info(f"{i}/{n_events} interacting = {np.sum(mask_int)}, failed = {failed}, n_cylinder = {n_cylinder}, eta = {pretty_time_delta(eta)}")
    #     print(f"calculating interaction point of event {i}"),
        c, s = np.cos(az[i]), np.sin(az[i])
        Raz = np.array(((c, -s, 0), (s, c, 0), (0, 0, 1)))
        c, s = np.cos(zen[i]), np.sin(zen[i])
    #     Rzen = np.array(((c, 1, -s), (0, 1, 0), (s, 0, c)))
        Rzen = hp.get_rotation(hp.spherical_to_cartesian(0, az[i]), hp.spherical_to_cartesian(zen[i], az[i]))
    #     R = hp.get_rotation(np.array([0, 0, 1]), hp.spherical_to_cartesian(zen[i], az[i]))
        R = np.matmul(Rzen, Raz)
        v = -hp.spherical_to_cartesian(zen[i], az[i])  # neutrino direction
        X = np.matmul(R, np.array([ax[i], ay[i], 0])) + np.array([0, 0, -0.5 * h_cylinder])
        if 0:
            import matplotlib.pyplot as plt
            from mpl_toolkits.mplot3d import Axes3D
            from mpl_toolkits.mplot3d.art3d import Poly3DCollection
            fig = plt.figure()
            a = fig.add_subplot(111, projection='3d')
            # Cylinder
            x = np.linspace(-r_cylinder, r_cylinder, 100)
            z = np.linspace(0, -h_cylinder, 100)
            Xc, Zc = np.meshgrid(x, z)
            Yc = np.sqrt(r_cylinder ** 2 - Xc ** 2)

            # Draw parameters
            rstride = 20
            cstride = 10
            a.plot_surface(Xc, Yc, Zc, alpha=0.2, rstride=rstride, cstride=cstride)
            a.plot_surface(Xc, -Yc, Zc, alpha=0.2, rstride=rstride, cstride=cstride)
            a.set_title(f"zenith = {zen[i]/units.deg:.0f}")
            xx = []
            yy = []
            zz = []
            for vert in np.array([[-0.5 * d, -0.5 * d, 0], [0.5 * d, -0.5 * d, 0], [0.5 * d, 0.5 * d, 0], [-0.5 * d, 0.5 * d, 0]]):
                t = np.matmul(Raz, vert) + np.array([0, 0, -0.5 * h_cylinder])
                xx.append(t[0])
                yy.append(t[1])
                zz.append(t[2])
            verts = [list(zip(xx, yy, zz))]
            a.add_collection3d(Poly3DCollection(verts, alpha=0.5))

            xx = []
            yy = []
            zz = []
            for vert in np.array([[-0.5 * d, -0.5 * d, 0], [0.5 * d, -0.5 * d, 0], [0.5 * d, 0.5 * d, 0], [-0.5 * d, 0.5 * d, 0]]):
                t = np.matmul(Rzen, np.matmul(Raz, vert)) + np.array([0, 0, -0.5 * h_cylinder])
                xx.append(t[0])
                yy.append(t[1])
                zz.append(t[2])
            verts = [list(zip(xx, yy, zz))]
            a.add_collection3d(Poly3DCollection(verts, alpha=0.5))

            s = np.array([0, 0, -0.5 * h_cylinder])
            t = v * 10 * units.km + s
            a.plot([s[0], t[0]], [s[1], t[1]], [s[2], t[2]], '-d')

            s = np.array([0, 0, -0.5 * h_cylinder])
            t = -hp.spherical_to_cartesian(90 * units.deg, az[i]) * 10 * units.km + s
            a.plot([s[0], t[0]], [s[1], t[1]], [s[2], t[2]], '--d')

            # check if neutrino axis is perpendicular
            t = np.array(verts[0][0]) - np.array(verts[0][1])
            t /= np.linalg.norm(t)
            print(np.dot(t, v))

            a.set_xlabel("x")
            a.set_zlabel("z")
            a.set_ylabel("y")
            plt.show()

        # check if trajectory passes through cylinder
    #     if(not points_in_cylinder(pt1, pt2, r_cylinder, X)):
        # we rotate everything in the plane defined by z and the propagration direction (such that v_y = 0)
        Xaz = np.matmul(Raz.T, X)
        rmin = Xaz[1]  # the closest distance to the z axis (center of cyllinder)
        if(abs(rmin) >= r_cylinder):
            continue
        # define the projected square of the cylinder
        # the two endpoints of the two horizontal lines are
        xtmp = (r_cylinder ** 2 - rmin ** 2) ** 0.5
        Lh1 = np.array([[-xtmp, 0], [xtmp, 0]])
        Lh2 = np.array([[-xtmp, -h_cylinder], [xtmp, -h_cylinder]])
        # the two endpoints of the two vertical lines are
        Lv1 = np.array([[-xtmp, 0], [-xtmp, -h_cylinder]])
        Lv2 = np.array([[xtmp, 0], [xtmp, -h_cylinder]])

        # define line of neutrino propagation by two points
        vaz = np.matmul(Raz.T, v)
        if(abs(vaz[1]) > 1e-10):
            a = 1 / 0
        v2d = np.array([vaz[0], vaz[2]])
        X2d = np.array([Xaz[0], Xaz[2]])
        t = 2 * d
        Paz = np.array([X2d + -t * v2d, X2d + t * v2d])

        # calculate points that intersect any of the 4 area (projected cylinder) boundaries
        intersects = []
        for k, (a1, a2) in enumerate([Lh1, Lh2]):
            tmp = seg_intersect(a1, a2, Paz[0], Paz[1])
            if((tmp[0] >= a1[0]) and (tmp[0] <= a2[0])):
                intersects.append(tmp)
        for k, (a1, a2) in enumerate([Lv1, Lv2]):
            tmp = seg_intersect(a1, a2, Paz[0], Paz[1])
            if((tmp[1] <= a1[1]) and (tmp[1] >= a2[1])):
                intersects.append(tmp)
        intersects = np.array(intersects)
        if(len(intersects) != 2):
            if 0:
                print(len(intersects))
                import matplotlib.pyplot as plt
                from mpl_toolkits.mplot3d import Axes3D
                fig = plt.figure()
                a = fig.add_subplot(111, projection='3d')
                # Cylinder
                x = np.linspace(-r_cylinder, r_cylinder, 100)
                z = np.linspace(0, -h_cylinder, 100)
                Xc, Zc = np.meshgrid(x, z)
                Yc = np.sqrt(r_cylinder ** 2 - Xc ** 2)

                # Draw parameters
                rstride = 20
                cstride = 10
                a.plot_surface(Xc, Yc, Zc, alpha=0.2, rstride=rstride, cstride=cstride)
                a.plot_surface(Xc, -Yc, Zc, alpha=0.2, rstride=rstride, cstride=cstride)
                X_enter = X + 10 * units.km * v
                X_leave = X - 10 * units.km * v
                a.plot([X_enter[0], X_leave[0]], [X_enter[1], X_leave[1]], [X_enter[2], X_leave[2]], '-o')

                a.set_xlabel("x")
                a.set_zlabel("z")
                a.set_ylabel("y")
                a.legend()
                a.set_title("no intersection")

                fig = plt.figure()
                a = fig.add_subplot(111, projection='3d')
                for (a1, a2) in [Lh1, Lh2, Lv1, Lv2]:
                    a.plot([a1[0], a2[0]], [rmin, rmin], [a1[1], a2[1]])
                a.plot([Paz[0][0], Paz[1][0]], [rmin, rmin], [Paz[0][1], Paz[1][1]])
                a.set_xlabel("x")
                a.set_zlabel("z")
                a.set_ylabel("y")
                a.legend()
                plt.show()
            continue  # neutrino is not passing through cylinder
        n_cylinder += 1
        ss = []
        for tmp in intersects:
            ss.append(np.dot(tmp - X2d, v2d.T))
        argsort = np.argsort(np.array(ss))  # check which intersection happens first along neutrino path
        if 0:
            if(len(intersects)):
                import matplotlib.pyplot as plt
                from mpl_toolkits.mplot3d import Axes3D
                fig = plt.figure()
                a = fig.add_subplot(111, projection='3d')
                for (a1, a2) in [Lh1, Lh2, Lv1, Lv2]:
                    a.plot([a1[0], a2[0]], [rmin, rmin], [a1[1], a2[1]])
                a.plot([Paz[0][0], Paz[1][0]], [rmin, rmin], [Paz[0][1], Paz[1][1]])
                for k, tmp in enumerate(intersects):
                    a.plot([tmp[0]], [rmin], [tmp[1]], 'o', label=f"s = {ss[k]:.0f}")
                a.set_xlabel("x")
                a.set_zlabel("z")
                a.set_ylabel("y")
                a.legend()
    #             plt.ion()
                plt.show()

        # calculate the 3D points where the neutrino enters/leaves the cylinder and transform to outside Earth
        X_enter = np.matmul(Raz, np.array([intersects[argsort][0][0], rmin, intersects[argsort][0][1]])) + np.array([0, 0, R_earth])
        X_leave = np.matmul(Raz, np.array([intersects[argsort][1][0], rmin, intersects[argsort][1][1]])) + np.array([0, 0, R_earth])
        X += np.array([0, 0, R_earth])

        # calculate point where neutrino enters Earth
        if(np.linalg.norm(X_enter) > R_earth):  # if enter point is outside of Earth (can happen because cylinder does not account for Earth curvature)
            if(np.linalg.norm(X_leave) > R_earth):  # check if leave point is also outside of Earth (can also happen because cylinder does not account for Earth curvature)
                continue
            t = brentq(obj_dist_to_surface, 0, 5 * d, args=(-v, X_leave))
            enter_point = X_leave + (-v * t)
            X_enter = enter_point  # define point where neutrino enters the cylinder as the point where it enters the Earth
        else:
            t = brentq(obj_dist_to_surface, 0, 2 * R_earth, args=(-v, X_enter))
            enter_point = X_enter + (-v * t)
    #     logger.debug(f"zen = {zen[i]/units.deg:.0f}deg, trajectory enters Earth at {enter_point[0]:.1f}, {enter_point[0]:.1f}, {enter_point[0]:.1f}. Dist to core = {np.linalg.norm(enter_point)/R_earth:.5f}, dist to (0,0,R) = {np.linalg.norm(enter_point - np.array([0,0,R_earth]))/R_earth:.4f}")

        # check if event interacts at all
        # calcualte slant depth to point of entering cylinder
        t = np.linalg.norm(enter_point - X_enter)
        slant_depth_min = slant_depth(t, v, enter_point)
        if(t == 0):
            slant_depth_min = 0
        # calculate slant depth through the cylinder
        s = np.linalg.norm(X_leave - X_enter)
        slant_depth_max = slant_depth(s, v, X_enter) + slant_depth_min  # full slant depth from outside Earth to point when it leaves the cylinder

        if 0:
            import matplotlib.pyplot as plt
            from mpl_toolkits.mplot3d import Axes3D
            fig = plt.figure()
            a = fig.add_subplot(111, projection='3d')
            # Cylinder
            x = np.linspace(-r_cylinder, r_cylinder, 100)
            z = np.linspace(0, -h_cylinder, 100)
            Xc, Zc = np.meshgrid(x, z)
            Yc = np.sqrt(r_cylinder ** 2 - Xc ** 2)

            # Draw parameters
            rstride = 20
            cstride = 10
            a.plot_surface(Xc, Yc, Zc, alpha=0.2, rstride=rstride, cstride=cstride)
            a.plot_surface(Xc, -Yc, Zc, alpha=0.2, rstride=rstride, cstride=cstride)
            a.plot([X_enter[0], X_leave[0]], [X_enter[1], X_leave[1]], [X_enter[2] - R_earth, X_leave[2] - R_earth], '-o')

            a.set_xlabel("x")
            a.set_zlabel("z")
            a.set_ylabel("y")
            print(f"Lmin = {slant_depth_min:.2g}, Lmax = {slant_depth_max:.2g}, Lnu = {Lint[i]:.2g}")
            a.legend()
            plt.show()
    #         a = 1 / 0
        if((Lint[i] <= slant_depth_min) or (Lint[i] >= slant_depth_max)):
            logger.debug("neutrino does not interact in cylinder, skipping to next event")
            continue

        try:
            # calculate interaction point by inegrating the density of Earth along the neutrino path until we wind the interaction length
            t = brentq(obj, 0, s, args=(v, X_enter, Lint[i] - slant_depth_min), maxiter=500)
        except:
            logger.warning("failed to converge, skipping event")
            failed += 1
            continue
        Xint = X_enter + v * t  # calculate interaction point

        is_in_cylinder = points_in_cylinder(pt1, pt2, r_cylinder, Xint)
        mask_int[i] = is_in_cylinder
        if(is_in_cylinder):
            logger.debug(f"event {i}, interaction point ({Xint[0]:.1f}, {Xint[1]:.1f}, {Xint[2]-R_earth:.1f}), in cylinder {is_in_cylinder}")
            data_sets['xx'].append(Xint[0])
            data_sets['yy'].append(Xint[1])
            data_sets['zz'].append(Xint[2] - R_earth)
            data_sets['zeniths'].append(zen[i])
            data_sets['azimuths'].append(az[i])
            data_sets['flavors'].append(flavors[i])
            data_sets['energies'].append(Enu[i])

        else:
            logger.error("interaction is not in cylinder but it should be")
            a = 1 / 0

    data_sets['event_ids'] = range(np.sum(mask_int))
    data_sets['flavors'] = np.ones(np.sum(mask_int))
    data_sets["event_ids"] = np.arange(np.sum(mask_int)) + start_event_id
    data_sets["n_interaction"] = np.ones(np.sum(mask_int), dtype=np.int)
    data_sets["vertex_times"] = np.zeros(np.sum(mask_int), dtype=np.float)

    data_sets["interaction_type"] = inelasticities.get_ccnc(np.sum(mask_int))
    data_sets["inelasticity"] = inelasticities.get_neutrino_inelasticity(np.sum(mask_int))

    attributes['n_events'] = n_events

    logger.info(f"{np.sum(mask_int)} event interacted in simulation volume")

    write_events_to_hdf5(filename, data_sets, attributes,
                         n_events_per_file=n_events_per_file, start_file_id=start_file_id)


if __name__ == "__main__":
    generate_eventlist_cylinder("test2.hdf5", 1e4, 1e18 * units.eV, 1e18 * units.eV, full_rmin=0, full_rmax=5 * units.km,
                                full_zmin=-2.7 * units.km, full_zmax=0, thetamin=0, thetamax=180 * units.deg,
                                phimin=0, phimax=360 * units.deg, start_event_id=0)



import os
import argparse
import logging
import math
from six import iteritems
import numpy as np
import h5py
logger = logging.getLogger("HDF5-split")
logging.basicConfig(level=logging.DEBUG)
logger.setLevel(logging.WARNING)


def split_hdf5_input_file(input_filename, output_filename, number_of_events_per_file):
    """
    splits up an existing hdf5 file into multiple subfiles

    Parameters
    ----------
    input_filename: string
        the input filename
    output_filename: string
        the desired output filename (if multiple files are generated, a 'part000x' is appended to the filename
    n_events_per_file: int (optional, default None)
        the number of events per file
    """
    fin = h5py.File(input_filename, 'r')
    attributes = {}
    data = {}
    groups = {}
    n_groups = {}
    n_data = {}
    group_attrs = {}

    for key in fin:
        if isinstance(fin[key], h5py._hl.group.Group):
            groups[key] = {}
            if(key not in n_groups):
                n_groups[key] = {}
            for key2 in fin[key]:
                groups[key][key2] = fin[key][key2][...]
                if(key2 not in n_groups[key]):
                    n_groups[key][key2] = 0
                n_groups[key][key2] += len(groups[key][key2])
            if(key not in group_attrs):
                group_attrs[key] = {}
                for key2 in fin[key].attrs:
                    group_attrs[key][key2] = fin[key].attrs[key2]
            else:
                for key2 in fin[key].attrs:
                    if(not np.all(group_attrs[key][key2] == fin[key].attrs[key2])):
                        logger.warning(f"attribute {key2} of group {key} of file {input_filename} are different ({group_attrs[key][key2]} vs. {fin[key].attrs[key2]}. Using attribute value of first file, but you have been warned!")
        else:
            data[key] = fin[key][...]
            if(key not in n_data):
                n_data[key] = 0
            n_data[key] += len(data[key])

    for key, value in iteritems(fin.attrs):
        attributes[key] = value
#     logger.info(f"setting number of events from {attributes['n_events']} to the actual number of events in this file {len(data_sets['event_ids'])}")
    fin.close()

    n_events = len(data['event_ids'])
    logger.info("saving {} events in total".format(n_events))
    total_number_of_events = attributes['n_events']

    n_files = math.ceil(n_events / number_of_events_per_file)
    for iFile in range(n_files):
        filename2 = output_filename + ".part{:04}".format(iFile)
        logger.debug(f"saving file {iFile} with {number_of_events_per_file} events to {filename2}")
        fout = h5py.File(filename2, 'w')
        for key, value in attributes.items():
            fout.attrs[key] = value
        fout.attrs['total_number_of_events'] = total_number_of_events / n_files
        fout.attrs['n_events'] = number_of_events_per_file

        for key, value in data.items():
            if value.dtype.kind == 'U':
                fout[key] = [np.char.encode(c, 'utf8') for c in value[iFile * number_of_events_per_file:(iFile + 1) * number_of_events_per_file]]
            else:
                fout[key] = value[iFile * number_of_events_per_file:(iFile + 1) * number_of_events_per_file]

        i1, i2 = iFile * number_of_events_per_file, (iFile + 1) * number_of_events_per_file
        for key in groups:
            logger.info("writing group {}".format(key))
            g = fout.create_group(key)
            for key2 in groups[key]:
                logger.info("writing data set {}".format(key2))

                shape = list(groups[key][key2][i1:i2].shape)
#                 shape[0] = n_groups[key][key2]

                tmp = groups[key][key2][i1:i2]

                g.create_dataset(key2, shape, dtype=groups[key][key2].dtype,
                                 compression='gzip')[...] = tmp
            # save group attributes
            for key2 in group_attrs[key]:
                fout[key].attrs[key2] = group_attrs[key][key2]

        fout.close()


if __name__ == "__main__":
    """
    Splits up a hdf5 file into multiple files
    
    Parameters
    -----------
    file: str
        the input file
    outputfolder: str
        the ouput folder
    n_events: int
        the maximum number of events in each file. The last file will contain less events. 
    Optional log level setting to either set DEBUG, INFO, or WARNING to the readout. Example: add --loglevel DEBUG when calling script to set loglevel to DEBUG. 
    """
    parser = argparse.ArgumentParser(description='Merge hdf5 files')
    parser.add_argument('file', type=str, help='input file')
    parser.add_argument('outputfolder', type=str, help='output folder')
    parser.add_argument('n_events', type=int, help='number of events per file')
    parser.add_argument('--loglevel', metavar='level', help='loglevel set to either DEBUG, INFO, or WARNING')
    args = parser.parse_args()

    if args.loglevel is not None:
        log_val = eval(f'logging.{args.loglevel}')
        logger.setLevel(log_val)

    if(not os.path.exists(args.outputfolder)):
        os.makedirs(args.outputfolder)
    input_filename = os.path.basename(args.file)

    split_hdf5_input_file(args.file, os.path.join(args.outputfolder, input_filename), args.n_events)



#!/usr/bin/env python
import numpy as np
import h5py
from NuRadioReco.utilities import units
import sys
import os

###########################
# Reference values from previous run, have to be updated, if code changes
###########################

# the event generation has a fixed seed and I switched to Alvarez2000 (also no randomness)
# thus, the Veff has no statistical scatter
Veff_mean = 5.45361
Veff_sigma = 0.0001

path = os.path.dirname(os.path.abspath(__file__))
fin = h5py.File(os.path.join(path, "output.hdf5"), 'r')


def calculate_veff(fin):
    weights = np.array(fin['weights'])
    n_events = fin.attrs['n_events']

    n_triggered = np.sum(weights)

    V = None
    rmin = fin.attrs['rmin']
    rmax = fin.attrs['rmax']
    dZ = fin.attrs['zmax'] - fin.attrs['zmin']
    V = np.pi * (rmax ** 2 - rmin ** 2) * dZ
    Veff = V * 4 * np.pi * np.sum(weights) / n_events
    return n_triggered, n_events, Veff

###########################
# calculate effective volume
###########################


n_triggered, n_events, Veff = calculate_veff(fin)

print('fraction of triggered events = {:.0f}/{:.0f} = {:.3f} -> {:.6g} km^3 sr'.format(n_triggered, n_events, n_triggered / n_events, Veff / units.km ** 3))

delta = (Veff / units.km ** 3 - Veff_mean) / Veff_sigma
rdelta = (Veff / units.km ** 3 - Veff_mean) / Veff_mean
print("effective volume deviates {:.1f} sigma ({:.0f}%) from the mean".format(delta, 100 * rdelta))

if(np.abs(Veff / units.km ** 3 - Veff_mean) > 3 * Veff_sigma):
    print("deviation is more than 3 sigma -> this should only happen in less than 1\% of the tests. Rerun the test and see if the error persists.")
    sys.exit(-1)

###########################
# Code to generate new average values for this test
###########################

Veffs = []

if False:
    files = os.listdir('.')
    for file in files:
        if (".hdf5" in file) and ("output" in file):
            fin = h5py.File(file, 'r')
            n_triggered, n_events, Veff = calculate_veff(fin)
            Veffs.append(Veff / units.km ** 3)
            print(Veff / units.km ** 3)

    print("New mean Veff {}".format(np.mean(Veffs)))
    print("New sigma Veff {}".format(np.std(Veffs)))




import numpy as np
from NuRadioReco.utilities import units
from scipy.interpolate import interp1d
from scipy import constants


def param(energy, inttype='cc'):
    """
    Parameterization and constants as used in
    get_nu_cross_section()
    See documentation there for details

    """

    if inttype == 'cc':
        c = (-1.826, -17.31, -6.406, 1.431, -17.91)  # nu, CC
    elif inttype == 'nc':
        c = (-1.826, -17.31, -6.448, 1.431, -18.61)  # nu, NC
    elif inttype == 'cc_bar':
        c = (-1.033, -15.95, -7.247, 1.569, -17.72)  # nu_bar, CC
    elif inttype == 'nc_bar':
        c = (-1.033, -15.95, -7.296, 1.569, -18.30)  # nu_bar, NC
    else:
        logger.error("Type {0} of interaction not defined".format(inttype))
        raise NotImplementedError

    epsilon = np.log10(energy / units.GeV)
    l_eps = np.log(epsilon - c[0])
    crscn = c[1] + c[2] * l_eps + c[3] * l_eps ** 2 + c[4] / l_eps
    crscn = np.power(10, crscn) * units.cm ** 2
    return crscn


def csms(energy, inttype, flavors):
    """
    Neutrino cross sections according to
    Amanda Cooper-Sarkar, Philipp Mertsch, Subir Sarkar
    JHEP 08 (2011) 042
    """
    if type(inttype) == str:
        inttype = np.array([inttype] * energy.shape[0])

    if (type(flavors) == int or type(flavors) == np.int64):
        flavors = np.array([flavors] * energy.shape[0])

    neutrino = np.array((
        [50, 0.32, 0.10],
        [100, 0.65, 0.20],
        [200, 1.3, 0.41],
        [500, 3.2, 1.0],
        [1000, 6.2, 2.0],
        [2000, 12., 3.8],
        [5000, 27., 8.6],
        [10000, 47., 15.],
        [20000, 77., 26.],
        [50000, 140., 49.],
        [100000, 210., 75.],
        [200000, 310., 110.],
        [500000, 490., 180.],
        [1e6, 690., 260.],
        [2e6, 950., 360.],
        [5e6, 1400., 540.],
        [1e7, 1900., 730.],
        [2e7, 2600., 980.],
        [5e7, 3700., 1400.],
        [1e8, 4800., 1900.],
        [2e8, 6200., 2400.],
        [5e8, 8700., 3400.],
        [1e9, 11000., 4400.],
        [2e9, 14000., 5600.],
        [5e9, 19000., 7600.],
        [1e10, 24000., 9600.],
        [2e10, 30000., 12000.],
        [5e10, 39000., 16000.],
        [1e11, 48000., 20000.],
        [2e11, 59000., 24000.],
        [5e11, 75000., 31000.]
        ))

    neutrino[:, 0] *= units.GeV
    neutrino[:, 1] *= units.picobarn  # CC
    neutrino[:, 2] *= units.picobarn  # NC

    neutrino_cc = interp1d(neutrino[:, 0], neutrino[:, 1], bounds_error=True)
    neutrino_nc = interp1d(neutrino[:, 0], neutrino[:, 2], bounds_error=True)

    antineutrino = np.array((
        [50, 0.15, 0.05],
        [100, 0.33, 0.12],
        [200, 0.69, 0.24],
        [500, 1.8, 0.61],
        [1000, 3.6, 1.20],
        [2000, 7., 2.4],
        [5000, 17., 5.8],
        [10000, 31., 11.],
        [20000, 55., 19.],
        [50000, 110., 39.],
        [100000, 180., 64.],
        [200000, 270., 99.],
        [500000, 460., 170.],
        [1e6, 660., 240.],
        [2e6, 920., 350.],
        [5e6, 1400., 530.],
        [1e7, 1900., 730.],
        [2e7, 2500., 980.],
        [5e7, 3700., 1400.],
        [1e8, 4800., 1900.],
        [2e8, 6200., 2400.],
        [5e8, 8700., 3400.],
        [1e9, 11000., 4400.],
        [2e9, 14000., 5600.],
        [5e9, 19000., 7600.],
        [1e10, 24000., 9600.],
        [2e10, 30000., 12000.],
        [5e10, 39000., 16000.],
        [1e11, 48000., 20000.],
        [2e11, 59000., 24000.],
        [5e11, 75000., 31000.]
    ))

    antineutrino[:, 0] *= units.GeV
    antineutrino[:, 1] *= units.picobarn  # CC
    antineutrino[:, 2] *= units.picobarn  # NC

    antineutrino_cc = interp1d(antineutrino[:, 0], antineutrino[:, 1], bounds_error=True)
    antineutrino_nc = interp1d(antineutrino[:, 0], antineutrino[:, 2], bounds_error=True)

    crscn = np.zeros_like(energy)

    particles_cc = np.where((flavors >= 0) & (inttype == 'cc'))
    particles_nc = np.where((flavors >= 0) & (inttype == 'nc'))
    antiparticles_cc = np.where((flavors < 0) & (inttype == 'cc'))
    antiparticles_nc = np.where((flavors < 0) & (inttype == 'nc'))
#
    crscn[particles_cc] = neutrino_cc(energy[particles_cc])
    crscn[particles_nc] = neutrino_nc(energy[particles_nc])
    crscn[antiparticles_cc] = antineutrino_cc(energy[antiparticles_cc])
    crscn[antiparticles_nc] = antineutrino_nc(energy[antiparticles_nc])

    return crscn


def get_nu_cross_section(energy, flavors, inttype='total', cross_section_type='ctw'):
    """
    return neutrino cross-section

    Parameters
        ----------
    energy: float/ array of floats
        neutrino energies/momenta in standard units

    flavors: float / array of floats
        neutrino flavor (integer) encoded as using PDG numbering scheme,
        particles have positive sign, anti-particles have negative sign, relevant are:
        12: electron neutrino
        14: muon neutrino
        16: tau neutrino

    inttype: str, array of str
        interaction type
        nc : neutral current
        cc : charged current
        total: total (for non-array type)

    cross_section_type: str
        defines model of cross-section
        ghandi : according to Ghandi et al. Phys.Rev.D58:093009,1998
                 only one cross-section for all interactions and flavors
        ctw    : A. Connolly, R. S. Thorne, and D. Waters, Phys. Rev.D 83, 113009 (2011).
                 cross-sections for all interaction types and flavors
        csms   : A. Cooper-Sarkar, P. Mertsch, S. Sarkar, JHEP 08 (2011) 042
    """

    if cross_section_type == 'ghandi':
        crscn = 7.84e-36 * units.cm ** 2 * np.power(energy / units.GeV, 0.363)

    elif cross_section_type == 'ctw':
        crscn = np.zeros_like(energy)
        if type(inttype) == str:
            if inttype == 'total':

                if (type(flavors) == int or type(flavors) == np.int64):
                    if flavors >= 0:
                        crscn = param(energy, 'nc') + param(energy, 'cc')
                    else:
                        crscn = param(energy, 'nc_bar') + param(energy, 'cc_bar')
                else:
                    antiparticles = np.where(flavors < 0)
                    particles = np.where(flavors >= 0)

                    crscn[particles] = param(energy[particles], 'nc') + param(energy[particles], 'cc')
                    crscn[antiparticles] = param(energy[antiparticles], 'nc_bar') + param(energy[antiparticles], 'cc_bar')

            if (inttype == 'cc') or (inttype == 'nc'):
                if (type(flavors) == int or type(flavors) == np.int64):
                    crscn = param(energy, inttype)
                else:
                    antiparticles = np.where(flavors < 0)
                    particles = np.where(flavors >= 0)
                    crscn[particles] = param(energy[particles], inttype)
                    crscn[antiparticles] = param(energy[antiparticles], inttype)
        else:

                if (type(flavors) == int or type(flavors) == np.int64):

                    particles_cc = np.where(inttype == 'cc')
                    particles_nc = np.where(inttype == 'nc')
                    if flavors >= 0:
                        crscn[particles_cc] = param(energy[particles_cc], 'cc')
                        crscn[particles_nc] = param(energy[particles_nc], 'nc')
                    else:
                        crscn[particles_cc] = param(energy[particles_cc], 'cc_bar')
                        crscn[particles_nc] = param(energy[particles_nc], 'nc_bar')

                else:
                    particles_cc = np.where((flavors >= 0) & (inttype == 'cc'))
                    particles_nc = np.where((flavors >= 0) & (inttype == 'nc'))
                    antiparticles_cc = np.where((flavors < 0) & (inttype == 'cc'))
                    antiparticles_nc = np.where((flavors < 0) & (inttype == 'nc'))

                    crscn[particles_cc] = param(energy[particles_cc], 'cc')
                    crscn[particles_nc] = param(energy[particles_nc], 'nc')
                    crscn[antiparticles_cc] = param(energy[antiparticles_cc], 'cc_bar')
                    crscn[antiparticles_nc] = param(energy[antiparticles_nc], 'nc_bar')

    elif cross_section_type == 'csms':
        crscn = csms(energy, inttype, flavors)

    else:
        logger.error("Cross-section {} not defined".format(cross_section_type))
        raise NotImplementedError

    return crscn


def get_interaction_length(Enu, density=.917 * units.g / units.cm ** 3, flavor=12, inttype='total',
                           cross_section_type='ctw'):
    """
    calculates interaction length from cross section

    Parameters
    ----------
    Enu: float
        neutrino energy
    density: float (optional)
        density of the medium, default density of ice = 0.917 g/cm**3
    flavors: float / array of floats
        neutrino flavor (integer) encoded as using PDG numbering scheme,
        particles have positive sign, anti-particles have negative sign, relevant are:
        12: electron neutrino
        14: muon neutrino
        16: tau neutrino

    inttype: str, array of str
        interaction type
        nc : neutral current
        cc : charged current
        total: total (for non-array type)

    cross_section_type: str
        defines model of cross-section
        ghandi : according to Ghandi et al. Phys.Rev.D58:093009,1998
                 only one cross-section for all interactions and flavors
        ctw    : A. Connolly, R. S. Thorne, and D. Waters, Phys. Rev.D 83, 113009 (2011).
                 cross-sections for all interaction types and flavors
        csms   : A. Cooper-Sarkar, P. Mertsch, S. Sarkar, JHEP 08 (2011) 042

    Returns float: interaction length

    """
    m_n = constants.m_p * units.kg  # nucleon mass, assuming proton mass
    L_int = m_n / get_nu_cross_section(Enu, flavors=flavor, inttype=inttype, cross_section_type=cross_section_type) / density
    return L_int


if __name__ == "__main__":  # this part of the code gets only executed it the script is directly called

    inttype = np.array(['cc'] * 10)

#     inttype = 'cc'

    flavors = np.array([14] * 10)

#     flavors = 14

    energy = np.array([1e12, 1e14, 1e16, 2e17, 6e17, 1e18, 5e18, 9e18, 1e19, 1e20]) * units.eV

    cc = get_nu_cross_section(energy, flavors, inttype=inttype) / units.picobarn

    cc_new = get_nu_cross_section(energy, flavors, inttype=inttype, cross_section_type='csms') / units.picobarn

    import matplotlib.pyplot as plt

    plt.figure()
    plt.loglog(energy / units.GeV, cc, label='CTW')
    plt.loglog(energy / units.GeV, cc_new, label='CSMS')
    plt.xlabel("Energy [GeV]")
    plt.ylabel("CC [pb]")
    plt.legend()

    plt.show()
